import time
from functools import lru_cache
from urllib.error import HTTPError

import numpy as np
import pandas as pd
import requests
import sidekick as sk

import mundi
from mundi import transforms
from ..cache import ttl_cache
from ..logging import log
from ..utils import today

HOURS = 3600
TIMEOUT = 6 * HOURS
EPIDEMIC_CURVES_APIS = {}
MOBILITY_DATA_APIS = {}


def epidemic_curve_api(key):
    return lambda fn: EPIDEMIC_CURVES_APIS.setdefault(key, fn)


def mobility_data_api(key):
    return lambda fn: MOBILITY_DATA_APIS.setdefault(key, fn)


#
# Epidemic curves
#
def epidemic_curve(region, api="auto", extra=False, **kwargs):
    """
    Universal interface to all epidemic curve loaders.

    Always return a dataframe with ["cases", "deaths"] columns for the given
    region. Some API's may offer additional columns such as "recovered", "test"
    etc.
    """
    code = mundi.code(region)
    fn = EPIDEMIC_CURVES_APIS[api]
    data = fn(code, **kwargs)
    return data if extra else data[["cases", "deaths"]]


@epidemic_curve_api("auto")
def auto_api(code, **kwargs):
    """
    Select best API to load according to region code.
    """
    if code == "BR" or code.startswith("BR-"):
        return brasil_io(code)
    elif len(code) == 2:
        return corona_api(code, **kwargs)
    raise ValueError(f"no API can load region with code: {code}")


@epidemic_curve_api("corona-api.com")
@ttl_cache("covid-19", timeout=TIMEOUT)
def corona_api(code) -> pd.DataFrame:
    """
    Load country's cases, deaths and recovered timeline from corona-api.com.
    """

    data = download_corona_api(code)
    data = data["data"]["timeline"]
    df = pd.DataFrame(data).rename({"confirmed": "cases"}, axis=1)

    df = df[["date", "cases", "deaths", "recovered"]]
    df["date"] = pd.to_datetime(df["date"])
    df = df.drop_duplicates("date", keep="first").set_index("date")
    df = df[df.fillna(0).sum(1) > 0].sort_index()

    # Fill missing data with previous measurement
    start, end = df.index[[0, -1]]
    full_index = pd.to_datetime(np.arange((end - start).days), unit="D", origin=start)
    df = df.reindex(full_index).fillna(method="ffill")

    return df.astype(int)


@sk.retry(10, sleep=0.5)
def download_corona_api(code) -> dict:
    log.info(f"[api/corona-api] Downloading data from corona API ({code})")

    url = "http://corona-api.com/countries/{code}?include=timeline"
    response = requests.get(url.format(code=code))
    size = len(response.content) // 1024

    log.info(f"[api/corona-api] Download ended with {size} kb")
    return response.json()


@epidemic_curve_api("brasil.io")
def brasil_io(code):
    cases = brasil_io_cases()
    cases = cases[cases["id"] == code].drop(columns="id")
    cases = cases.drop_duplicates("date")
    return cases.set_index("date").sort_index()


@ttl_cache("covid-19", timeout=TIMEOUT)
def brasil_io_cases() -> pd.DataFrame:
    """
    Return the complete dataframe of cases and deaths from Brasil.io.
    """

    df = download_brasil_io_cases()
    cols = {
        "last_available_confirmed": "cases",
        "confirmed": "cases",
        "last_available_deaths": "deaths",
        "city_ibge_code": "code",
    }

    cases = df.rename(cols, axis=1)
    cases = cases[cases["code"].notna()]

    cases["code"] = cases["code"].apply(lambda x: str(int(x))).astype("string")
    cases["code"] = "BR-" + cases["code"]

    cases["date"] = pd.to_datetime(cases["date"])
    cases = cases[cases["place_type"] == "city"]

    cases = cases[["date", "code", "cases", "deaths"]]
    cases = cases.dropna().reset_index(drop=True)
    cases = cases.rename({"code": "id"}, axis=1)

    log.info(f"[api/brasil.io] Merging {len(df)} entries")

    result = {}
    for col in ["cases", "deaths"]:
        data = cases.pivot_table(index="id", columns="date", values=col).fillna(-1).sort_index()
        data = transforms.sum_children(data).reset_index()
        data = data.melt(id_vars=["id"], var_name="date", value_name=col)
        data = data[data[col] >= 0]
        result[col] = data
    out = (
        pd.merge(*result.values(), on=["id", "date"], how="outer")
        .fillna(0)
        .astype({"cases": int, "deaths": int})
    )

    log.info("[api/brasil.io] Merge complete")

    return out


@sk.retry(10, sleep=0.5)
def download_brasil_io_cases():
    log.info("[api/brasil.io] Downloading data from Brasil.io")

    url = "https://data.brasil.io/dataset/covid19/caso_full.csv.gz"
    try:
        return pd.read_csv(url)
    except HTTPError as e:
        log.warn(f"[api/brasil.io] error downloading: {e}, using Github fallback")
        url = "https://github.com/pydemic/databases/raw/master/caso_full.csv.gz"
        return pd.read_csv(url)


#
# Mobility data
#
@ttl_cache("covid-19", timeout=TIMEOUT)
@sk.retry(10, sleep=0.5)
def google_mobility_data(cli=False):
    """
    Download google mobility data
    """
    url = "https://www.gstatic.com/covid19/mobility/Global_Mobility_Report.csv"

    log.info(f"Downloading google mobility data {today()}")
    t0 = time.time()
    data = requests.get(url)
    log.info(f"Download finished after {time.time() - t0:0.2} seconds")

    data_cols = ["retail", "grocery", "parks", "transit", "work", "residential"]

    df = pd.read_csv(data.content.decode("utf8")).rename(
        {
            "retail_and_recreation_percent_change_from_baseline": "retail",
            "grocery_and_pharmacy_percent_change_from_baseline": "grocery",
            "parks_percent_change_from_baseline": "parks",
            "transit_stations_percent_change_from_baseline": "transit",
            "workplaces_percent_change_from_baseline": "work",
            "residential_percent_change_from_baseline": "residential",
        },
        axis=1,
    )
    df["date"] = pd.to_datetime(df["date"])
    df[data_cols] = df[data_cols] / 100.0
    return df


def fix_google_mobility_data_region_codes(df):
    data = df[["country_region_code", "sub_region_1", "sub_region_2"]]
    codes = data.apply(subregion_code)
    return df


@lru_cache(1024)
def subregion_code(country, region, subregion):
    region = region or None
    subregion = subregion or None

    # Check arbitrary mapping
    mapping = google_mobility_map_codes()
    try:
        return mapping[country, region, subregion]
    except KeyError:
        pass

    # Fasttrack pure-country codes
    if not region:
        return country

    for name in (subregion, region):
        try:
            region = mundi.region(country_code=country, name=name)
        except LookupError:
            return region.id

    return country + "-" + region


@lru_cache(1)
def google_mobility_map_codes() -> dict:
    data = {}

    # Brazilian states
    for state in mundi.regions("BR", type="state"):
        data["BR", f"State of {state}", None] = state.id
    data["BR", "Federal District", None] = "BR-DF"

    return data


if __name__ == "__main__":
    sk.import_later("..cli.api:covid19_api_downloader", package=__package__)()

    # df = google_mobility_data()
    # df = fix_google_mobility_data_region_codes(df)
    # print(df)



from abc import ABC
from collections import defaultdict, namedtuple
from functools import lru_cache
from gettext import gettext as _
from typing import TYPE_CHECKING

import pandas as pd

from pydemic.utils import fmt, pc
from .results import Results

if TYPE_CHECKING:
    from .meta_info import Meta
    from ..models import Model

SummaryRow = namedtuple("Row", ["name", "variable", "value", "default", "unit"])


class WithResultsMixin(ABC):
    """
    Subclasses that adopt this mixin gain a "summary" attribute that expose
    useful information about its owner.

    In the context of simulation models, it shows useful information about the
    evolution of the epidemic.
    """

    meta: "Meta"
    RESULT_DATES_KEYS = ("start", "end", "peak")

    @property
    def results(self) -> Results:
        return Results(self)

    def __init__(self):
        self._results_cache = defaultdict(dict)
        self._results_dirty_check = None

    def __getitem__(self, item):
        raise NotImplementedError

    #
    # Reporting and summarization
    #
    def summary_table(self: "Model", role="all", humanize=False, translate=_):
        """
        Return the summary table with model parameters.
        """

        if role == "mortality":
            params = self.disease_params
            tables = []

            try:
                tables.append(getattr(params, "mortality_table"))
            except AttributeError:
                pass

            try:
                tables.append(getattr(params, "hospitalization_table"))
            except AttributeError:
                pass

            df = pd.concat(tables, axis=1)
            df.columns = map(_, df.columns)
            df = df.applymap(pc)
            df.index = [*(f"{n}-{n + 9}" for n in df.index[:-1]), f"{df.index[-1]}+"]
            if humanize:
                rename = humanized_summary_cols(translate)
                df = df.rename(rename, axis=1)
                df.index.name = None
            return df

        if isinstance(role, str):
            role = summary_table_roles(translate)[role]

        default_params = self.disease.params()
        data = []
        for k, (name, unit, fn) in role.items():
            value = fn(getattr(self, k, None))
            default = fn(getattr(default_params, k, None))
            row = SummaryRow(name, k, value, default, unit)
            data.append(row)

        df = pd.DataFrame(data).set_index("variable")
        if humanize:
            rename = humanized_summary_cols(translate)
            df = df.rename(rename, axis=1)
            df.index.name = rename[df.index.name]

        return df

    #
    # Results methods
    #
    def get_results_keys_data(self):
        """
        Yield keys for the result["data"] dict.
        """
        yield from self.meta.variables
        yield "cases"
        yield "attack_rate"

    def get_results_value_data(self, key):
        """
        Return value for model.result["data.<key>"] queries.
        """
        if key == "attack_rate":
            population = self["population:initial"]
            return (population - self["susceptible:final"]) / population
        return self[f"{key}:final"]

    def get_results_keys_params(self):
        """
        Yield keys for the result["params"] dict.
        """
        yield from self.meta.params.primary

    def get_results_value_params(self: "Model", key):
        """
        Return value for model.result["params.<key>"] queries.
        """
        return self.get_param(key)

    def get_results_keys_dates(self):
        """
        Yield keys for the result["dates"] dict.
        """
        yield from self.RESULT_DATES_KEYS

    def get_results_value_dates(self: "Model", key):
        """
        Return value for model.result["dates.<key>"] queries.
        """
        if key == "start":
            return self.to_date(0)
        elif key == "end":
            return self.date
        elif key == "peak":
            return self["infectious:peak-date"]
        else:
            raise KeyError(key)


@lru_cache(8)
def humanized_summary_cols(translate=_):
    _ = translate
    return {
        "name": _("Name"),
        "value": _("Model value"),
        "default": _("Default value"),
        "unit": _("Unit"),
        "variable": _("Variable"),
        "severe": _("Severe"),
        "critical": _("Critical"),
    }


@lru_cache(8)
def summary_table_roles(translate=_):
    _ = translate
    epidemiological_params = {
        "R0": (_("R0"), "-", fmt),
        "duplication_time": (_("Duplication time"), _("days"), fmt),
        "infectious_period": (_("Infectious period"), _("days"), fmt),
        "incubation_period": (_("Incubation period"), _("days"), fmt),
        "prodromal_period": (_("Prodromal period"), _("days"), fmt),
        "rho": (_("Relative infectiousness"), "-", pc),
        "prob_symptoms": (_("Probability of developing symptoms"), "-", pc),
    }

    clinical_periods_params = {
        "severe_period": (_("Average hospitalization duration"), _("days"), fmt),
        "critical_period": (_("Average ICU duration"), _("days"), fmt),
        "symptoms_delay": (_("Infection to symptoms onset"), _("days"), fmt),
        "severe_delay": (_("Symptoms onset to severe"), _("days"), fmt),
        "critical_delay": (_("Symptoms onset to critical"), _("days"), fmt),
        "death_delay": (_("Symptoms onset to death"), _("days"), fmt),
    }

    clinical_rate_params = {
        "prob_severe": (_("Probability of developing severe symptoms"), "-", pc),
        "prob_critical": (_("Probability of developing critical symptoms"), "-", pc),
        "CFR": (_("Case fatality ratio"), "-", pc),
        "IFR": (_("Infection fatality ratio"), "-", pc),
        "HFR": (_("Hospitalization fatality ratio"), "-", pc),
        "ICUFR": (_("ICU fatality ratio"), "-", pc),
    }

    return {
        "epidemic": epidemiological_params,
        "clinical_periods": clinical_periods_params,
        "clinical_rates": clinical_rate_params,
        "clinical": {**clinical_rate_params, **clinical_periods_params},
        "all": {**epidemiological_params, **clinical_periods_params, **clinical_rate_params},
    }



import sidekick as sk
from sidekick import placeholder as _

from .model import ClinicalObserverModel
from .utils import delayed_with_discharge
from ..params import clinical
from ..utils import param_property, param_alias, sliced


class CrudeFR(ClinicalObserverModel):
    """
    Model in which infected can become hospitalized and suffer a constant
    hospitalization fatality rate.

    Attributes:
        prob_severe (float, alias: Qsv):
            Probability that regular cases become severe (cases that require
            hospitalization).
        prob_critical (float, alias: Qcr):
            Probability that regular cases become critical (require ICU).
        hospitalization_period:
            Average duration of hospitalizations.
        icu_period:
            Average duration of ICU treatment.
        hospital_fatality_ratio (float, alias: HFR):
            Fraction of deaths for patients that go to hospitalization.
        icu_fatality_ratio (float, alias: ICUFR):
            Fraction of deaths for patients that go to ICU treatment.
    """

    params = clinical.DEFAULT

    # Primary parameters
    prob_severe: float = param_property(default=0.0)
    prob_critical: float = param_property(default=0.0)

    severe_period: float = param_property(default=0.0)
    critical_period: float = param_property(default=0.0)

    # Aliases
    Qsv: float = param_alias("prob_severe")
    Qcr: float = param_alias("prob_critical")

    # Properties
    hospital_fatality_ratio = sk.property(_.CFR / _.Qsv)
    HFR = sk.alias("hospital_fatality_ratio")

    icu_fatality_ratio = sk.property(_.CFR / _.Qcr)
    ICUFR = sk.alias("icu_fatality_ratio")

    prob_aggravate_to_icu = sk.property(_.Qcr / _.Qsv)

    # Cumulative series
    def get_data_deaths(self, idx):
        return self["cases", idx] * self.CFR

    def get_data_severe(self, idx):
        data = self["severe_cases"]
        K = max(self.K, 0)
        data = delayed_with_discharge(data, 0, self.severe_period, K, positive=True)
        return sliced(data, idx)

    def get_data_severe_cases(self, idx):
        return self["cases", idx] * self.Qsv

    def get_data_critical_cases(self, idx):
        return self["cases", idx] * self.Qcr



from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
from sidekick import X

if TYPE_CHECKING:
    X: list
    Y: list
    V: list
    W: list


#
# Color handling
#
def color(i=1, ax=None):
    """
    Return the color of last i-th line in the current (or given) axis.

    This function is useful to repeat color of past plots.

    Examples:
        >>> plt.plot(Y)
        >>> plt.plot(W)
    """
    ax = ax or plt.gca()
    try:
        return ax.lines[-i].get_color()
    except IndexError:
        return None


#
# Lines
#
def mark_x(x, *args, text=None, ax=None, **kwargs):
    """
    Create a vertical line in the x position.

    Accepts all standard arguments for the plt.plot() function.

    Args:
        x (float):
            x position of the vertical line.
        text:
            Optional label written alongside the line.
        ax:
            Optional axes object.

    See Also:
        :func:`mark_y`
    """
    ax = ax or plt.gca()
    y0, y1 = ax.get_ylim()
    ax.plot([x, x], [y0, y1], *args, **kwargs)
    ax.set_ylim(y0, y1)


def mark_y(y, *args, text=None, ax=None, **kwargs):
    """
    Create a vertical line in the x position.

    Accepts all standard arguments for the plt.plot() function.

    Args:
       y (float):
           y position of the vertical line.
       text:
           Optional label written alongside the line.
       ax:
           Optional axes object.

    See Also:
        :func:`mark_x`
    """
    ax = ax or plt.gca()
    x0, x1 = ax.get_xlim()
    ax.plot([x0, x1], [y, y], *args, **kwargs)
    ax.set_xlim(x0, x1)


#
# Layout
#
def tight(which="both", ax=None):
    """
    Remove margins in either 'x', 'y' or 'both' directions.

    Examples:
        >>> plt.plot(Y)
        >>> tight('x')
    """
    if which == "x":
        x, y = True, False
    elif which == "y":
        x, y = False, True
    elif which == "both":
        x = y = True
    elif which == "none":
        return
    else:
        raise ValueError(f"invalid axis: {which!r}")

    ax = ax or plt.gca()
    lines = ax.lines

    if x:
        lines = [line.get_xdata()[[0, -1]] for line in lines]
        vmin = min(map((X[0]), lines))
        vmax = max(map((X[1]), lines))
        ax.set_xlim(vmin, vmax)
    if y:
        lines = [line.get_ydata()[[0, -1]] for line in lines]
        vmin = min(map((X[0]), lines))
        vmax = max(map((X[1]), lines))
        ax.set_ylim(vmin, vmax)


#
# Dates
#
# ... TODO



from typing import Union, TYPE_CHECKING

import sidekick as sk

if TYPE_CHECKING:
    from ..clinical_models import CrudeFR, HospitalizationWithDelay, HospitalizationWithOverflow

    ClinicalModel = Union["CrudeFR", "HospitalizationWithDelay", "HospitalizationWithOverflow"]


class Clinical:
    """
    Implements the ``.clinical`` attribute of models.
    """

    __slots__ = ("_model",)

    _default = sk.lazy(lambda x: x())
    info = sk.delegate_to("_default")
    params = sk.delegate_to("_default")
    summary = sk.delegate_to("_default")

    @property
    def _models(self):
        from .. import clinical_models

        return clinical_models

    def __init__(self, model):
        self._model = model

    def __call__(self, *args, **kwargs):
        params = {**self._model.clinical_params, **kwargs}
        if args:
            (cls,) = args
            if isinstance(cls, str):
                return self.clinical_model(cls, **params)
        else:
            cls = self._model.clinical_model or self._models.CrudeFR
        return self.clinical_model(cls, **params)

    def __getitem__(self, item):
        return self._default[item]

    def clinical_model(self, cls, **kwargs) -> "ClinicalModel":
        """
        Create a clinical model from model infectious model instance.
        """
        if isinstance(cls, str):
            method = getattr(self, cls.lower() + "_model", None)
            if method is None:
                raise ValueError(f"invalid model: {cls!r}")
            return method(**kwargs)

        return cls(self._model, **kwargs)

    def crude_model(self, **kwargs) -> "CrudeFR":
        """
        Create a clinical model from model infectious model instance.
        """
        cls = self._models.CrudeFR
        return self.clinical_model(cls, **kwargs)

    def delay_model(self, **kwargs) -> "HospitalizationWithDelay":
        """
        A simple clinical model in which hospitalization occurs with some
        fixed delay.
        """
        cls = self._models.HospitalizationWithDelay
        return self.clinical_model(cls, **kwargs)

    def overflow_model(self, **kwargs) -> "HospitalizationWithOverflow":
        """
        A clinical model that considers the overflow of a healthcare system
        in order to compute the total death toll.
        """
        cls = self._models.HospitalizationWithOverflow
        return self.clinical_model(cls, **kwargs)



import numpy
import pandas
import sidekick
from sidekick import import_later as _imp

plt = _imp("matplotlib.pyplot")
click = _imp("click")
sns = _imp("seaborn")
sm = _imp("statsmodels.api")
smf = _imp("statsmodels.formula.api")
integrate = _imp("scipy.integrate")

np = numpy
pd = pandas
sk = sidekick



from typing import TypeVar

import numpy as np
import pandas as pd

T = TypeVar("T")


def diff(data: T, prepend=0, append=None, smooth=False) -> T:
    """
    Compute difference and return an object with the same shape as the
    original.

    Args:
        data (DataFrame or array):
            Input data to take the derivative.
        prepend:
            Value before the first data point.
        append:
            Value after the last data point. If given, take forward differences
            of data.
        smooth (bool):
            If true, apply the :func:`smooth` function to data. If smooth is a
            number, it is treated as the window size in the triangular
            smoothing process.
    """

    if append is not None:
        out = np.diff(data, axis=0, append=append)
    else:
        out = np.diff(data, axis=0, prepend=prepend)

    if isinstance(data, pd.DataFrame):
        data = pd.DataFrame(out, columns=data.columns, index=data.index)
    elif isinstance(data, pd.Series):
        data = pd.Series(out, index=data.index)
    else:
        data = out

    if smooth:
        window = 14 if smooth is True else smooth
        return _smooth(data, window)
    return data


def smooth(data: T, window=14) -> T:
    """
    Smooth data using a triangular window with the given window size.
    """

    if not isinstance(data, (pd.Series, pd.DataFrame)):
        return smooth(pd.DataFrame(data), window).values

    return data.rolling(window, win_type="triang", min_periods=1, center=True).mean()


def cases(curves) -> pd.Series:
    """
    Return the "cases" column if the dataframe has multiple columns.
    """
    if isinstance(curves, pd.Series):
        return curves
    elif len(curves.columns) == 1:
        return curves.iloc[:, 0]
    else:
        return curves["cases"]


_smooth = smooth



"""
Implement interfaces for classes that expose lists of parameters.
"""
from abc import ABC
from numbers import Number
from typing import Union

import pandas as pd
import sidekick as sk
from sidekick import placeholder as _

from .params_info import ParamsInfo
from ..logging import log
from ..params import Param, param, get_param
from ..types import ImproperlyConfigured
from ..utils import extract_keys

param_ = param


class WithParamsMixin(ABC):
    """
    Basic interface for classes that have parameters
    """

    name: str
    params: ParamsInfo
    params_data: pd.DataFrame
    disease_params: object
    _params: dict

    # Cache information in the params_info object as instance attributes.
    __all_params: frozenset = sk.lazy(_.meta.params.all)
    __primary_params: frozenset = sk.lazy(_.meta.params.primary)
    __alternative_params: frozenset = sk.lazy(_.meta.params.alternative)

    def __init__(self, params=None, keywords=None):
        self._params = {}
        if params is not None:
            self.set_params(params)

        extra = self.__all_params.intersection(keywords)
        if extra:
            self.set_params(extract_keys(extra, keywords))

        for key in self.__primary_params - self._params.keys():
            self.set_param(key, init_param(key, self, self.disease_params))

        s1 = frozenset(self._params)
        s2 = self.__primary_params
        assert s1 == s2, f"Different param set: {s1} != {s2}"

    #
    # Parameters
    #
    def set_params(self, params=None, **kwargs):
        """
        Set a collection of params.
        """

        for k, v in kwargs:
            self.set_param(k, v)

        if params:
            NOT_GIVEN = object()
            for p in self.__primary_params:
                if p not in kwargs:
                    value = get_param(p, params, default=NOT_GIVEN)
                    if value is not NOT_GIVEN:
                        self._params[p] = param_(value)

    def set_param(self, name, value, *, pdf=None, ref=None):
        """
        Sets a parameter in the model, possibly assigning a distribution and
        reference.
        """
        if name in self.__primary_params:
            cls = type(self).__name__
            log.debug(f"{cls}.{name} = {value!r} ({self.name})")

            self._params[name] = param(value, pdf=pdf, ref=ref)
        elif name in self.__all_params:
            setattr(self, name, param(value).value)
        else:
            raise ValueError(f"{name} is an invalid param name")

    def get_param(self, name, param=False) -> Union[Number, Param]:
        """
        Return the parameter with given name.

        Args:
            name:
                Parameter name.
            param:
                If True, return a :cls:`Param` instance instead of a value.
        """
        if param:
            try:
                return self._params[name]
            except KeyError:
                return param_(self.get_param(name))
        try:
            return self._params[name].value
        except KeyError:
            pass
        if name in self.__all_params:
            return getattr(self, name)
        else:
            raise ValueError(f"invalid parameter name: {name!r}")


class WithDynamicParamsMixin(WithParamsMixin, ABC):
    """
    Like HasParams, but allow parameters to be set up as functions.
    """


def init_param(key, obj, disease):
    if disease is not None:
        try:
            return get_param(key, disease)
        except ValueError:
            pass
    try:
        return getattr(obj, key)
    except AttributeError:
        pass

    try:
        value = getattr(obj, "_" + key)
    except AttributeError:
        raise ImproperlyConfigured(
            f"""bad parameter: {key}

Classes must either provide explicit default values for parameters or
implement a _{key}() method used to initialize the parameter during instance
creation.

The initialization method may explicitly raise exceptions, if it is not
possible to provide a sane default value for the instance.
"""
        )
    else:
        if callable(value):
            return value()
        return value



import numpy as np

from .abstract_seair import AbstractSEAIR
from .ode_model import ODEModel
from ..packages import click


class SEAIR(ODEModel, AbstractSEAIR):
    """
    A simple SIR model linearized around the DFE.
    """

    def diff(self, x, t):
        s, e, a, i, r = x
        n = s + e + a + i + r
        beta = self.beta
        gamma = self.gamma
        sigma = self.sigma
        rho = self.rho
        Qs = self.Qs

        infections = beta * s * ((i + rho * a) / n)
        return np.array(
            [
                -infections,
                +infections - sigma * e,
                +(1 - Qs) * sigma * e - gamma * a,
                +Qs * sigma * e - gamma * i,
                +gamma * (i + a),
            ]
        )


if __name__ == "__main__":

    @click.command()
    @click.option("--R0", "-r", default=2.74, help="Reproduction number")
    @click.option("--duration", "-t", default=90, type=int, help="Duration")
    @click.option("--log", default=False, type=bool, help="Use log-scale?")
    def cli(duration, r0, log):
        m = SEAIR()
        m.R0 = r0
        m.run(duration)
        m.plot(logy=log, show=True)

    cli()



import pandas as pd

from ..utils import trim_weeks, accumulate_weekly


def weekday_rate(data):
    """
    Read time series data and show the fraction accumulated in each weekday
    from a series of observations.
    """

    data = trim_weeks(data)
    by_week = accumulate_weekly(data, "trim")

    mean = []
    std = []

    for weekday in range(7):
        values = data.iloc[weekday::7].reset_index(drop=True)
        ratios = values / by_week

        mean.append(ratios.mean())
        std.append(ratios.std())

    means = pd.DataFrame(mean)
    std = pd.DataFrame(std)
    std.columns = [col + "_std" for col in std.columns]

    return pd.concat([means, std], axis=1)



from pandas.testing import assert_series_equal, assert_frame_equal
from pytest import approx

import mundi
from pydemic.models import SIR
from pydemic.utils import flatten_dict


class TestInfoMixin:
    def test_model_to_json(self):
        m = SIR()

    def _test_regional_model_to_json(self):
        m = SIR(region="BR")
        br = mundi.region("BR")
        assert m.info.to_dict() == {
            "demography.population": br.population,
            "demography.age_distribution": br.age_distribution,
            "demography.age_pyramid": br.age_pyramid,
        }
        assert m.info.to_dict(flat=True) == flatten_dict(m.info.to_dict())


class TestRegionMixin:
    def test_initialization_options(self):
        br = mundi.region("BR")
        it = mundi.region("IT")

        # Initialize with no region
        m = SIR()
        assert m.population == 1_000_000
        assert m.region is None
        assert m.age_distribution is None
        assert m.age_pyramid is None

        # Use a region
        m = SIR(region="BR")
        assert m.population == br.population
        assert_series_equal(m.age_distribution, br.age_distribution, check_names=False)
        assert_frame_equal(m.age_pyramid, br.age_pyramid)

        # Mix parameters a region
        tol = 1e-6
        m = SIR(region="BR", population=1000)
        assert m.population == 1000
        assert abs(m.age_distribution.sum() - 1000) < tol
        assert abs(m.age_pyramid.sum().sum() - 1000) < tol

        ratio = br.age_distribution / m.age_distribution
        assert ((ratio - br.population / 1000).dropna().abs() < tol).all()

        # Mixed values: brazilian population with the age_distribution proportions
        # from Italy
        m = SIR(region="BR", age_distribution="IT")
        assert m.population == br.population
        assert m.age_distribution.sum() == approx(br.population)
        assert list(m.age_distribution / m.population) == approx(
            it.age_distribution / it.population
        )



import numpy as np

from .abstract_sir import AbstractSIR


class eSIR(AbstractSIR):
    """
    A simple SIR model linearized around the DFE.
    """

    def run_to_fill(self, data, ts):
        ts = ts - self.time

        s0 = self.susceptible
        i0 = self.infectious
        r0 = self.recovered
        n = s0 + i0 + r0
        e = max(s0 / n, 0.0)

        R0 = self.R0
        gamma = self.gamma
        Ke = gamma * (e * R0 - 1)

        # Recovered and susceptible
        i = i0 * np.exp(Ke * ts)
        if abs(Ke) < 1e-6:
            x = Ke * ts
            factor = i0 * gamma * ts * (1 + x / 2 + x * x / 2)
        else:
            factor = (gamma / Ke) * np.maximum(i - i0, 0.0)
        r = np.minimum(r0 + factor, n)
        s = np.maximum(s0 - R0 * e * factor, 0.0)

        # Save data
        data[:, 0] = s
        data[:, 1] = i
        data[:, 2] = r


if __name__ == "__main__":
    import click

    @click.command()
    @click.option("--R0", "-r", default=2.74, help="Reproduction number")
    @click.option("--duration", "-t", default=90, type=int, help="Duration")
    @click.option("--steps", "-s", default=10, type=int, help="Number of steps")
    @click.option("--log", default=False, type=bool, help="Use log-scale?")
    def cli(duration, steps, r0, log):
        m = eSIR()
        m.R0 = r0

        while steps:
            step = round(duration / steps)
            m.run(step)

            steps -= 1
            duration -= step

        m.plot(logy=log, show=True)

    cli()



import datetime
from functools import total_ordering
from numbers import Number
from typing import Iterator, Dict, TYPE_CHECKING

import sidekick as sk

from .base_attr import BaseAttr
from ..utils import fmt, as_seq

NOT_GIVEN = object()
if TYPE_CHECKING:
    pass


@total_ordering
class Event(sk.Record):
    """
    Represents an Event.
    """

    key: str
    time: int
    date: datetime.date
    description: str = ""
    tags: set = frozenset()

    @classmethod
    def from_model(cls, model, key, time_or_date, description=None, tags=()):
        """
        Create an event instance passing model and only a single day or date.
        """
        if isinstance(time_or_date, (float, int)):
            time = time_or_date
            date = model.to_date(time)
        else:
            date = time_or_date
            time = model.to_time(date)
        description = description or key.replace("_", " ").title()
        return Event(key, time, date, description=description, tags=set(tags))

    def __str__(self):
        return f"{self.description} ({fmt(self.date)})"

    def __eq__(self, other):
        if isinstance(other, Event):
            return self[:-1] == other[:-1]
        return NotImplemented

    def __gt__(self, other):
        if isinstance(other, Event):
            return self.time > other.time
        return NotImplemented

    def __ge__(self, other):
        if isinstance(other, Event):
            return self.time >= other.time
        return NotImplemented

    def tag(self, tags):
        """
        Tag event with the given tag or list of tags.

        Notes:
            The event tags set is a publicly exposed attribute and can be used
            to make queries such as the examples:

             * ``tag in ev.tags``: check if event has tag.
             * ``ev.tags.issuperset(seq)``: check if event has all tags in sequence.
             * ``ev.tags & set(seq)``: check tags has some tags in seq.
        """
        self.tags.update(as_seq(tags))


class Info(BaseAttr):
    """
    Info objects store static information about a simulation.

    Not all information exposed by this object is necessarily relevant to a
    simulation model, but might be useful for later analysis or for
    convenience.

    Info attributes are organized in a dotted namespace.
    """

    __slots__ = ()
    _method_namespace = "info"

    @property
    def _cache(self):
        # noinspection PyProtectedMember
        return self.model._info_cache

    def clear(self):
        self._cache.clear()

    def save_event(self, key, time_or_date=None, description=None):
        """
        Save event under the given key.

        Args:
            key:
                A unique identifier for the event.
            time_or_date:
                When the event happened. If not given, assumes the current
                instant.
            description:
                An optional description for the event.
        """
        if time_or_date is None:
            time_or_date = self.model.time
        event = Event.from_model(self.model, key, time_or_date, description)
        self._set_item("event", key, event)

    def get_events(self, *args, tags=False) -> Iterator[Event]:
        """
        Return all events that happened in the given day or date or interval.

        * info.get_events(day) -> events in a specific day
        * info.get_events(a, b) -> events between a and b (inclusive)
        * info.get_events(a, ...) -> events beginning at a
        * info.get_events(..., b) -> events up to b
        """
        events = self["events"]
        model = self.model

        if len(args) == 1:
            seq = select_events_at(events, model.to_time(args[0]))
        elif args == 2:
            args = map(model.to_time, args)
            seq = select_events_between(events, *args)
        else:
            raise TypeError("method requires 1 or 2 arguments")

        if tags:
            tags = set(tags)
            for ev in seq:
                if tags.issubset(ev.tags):
                    yield ev
        else:
            yield from seq

    def _set_item(self, group, key, value):
        if group == "event" and not isinstance(value, Event):
            if isinstance(value, (int, datetime.date)):
                value = Event.from_model(self.model, key, value)
            else:
                value = Event.from_model(self.model, key, *value)

        super()._set_item(group, key, value)


def select_events_at(evs: Dict[str, Event], a: Number) -> Iterator[Event]:
    return (ev for k, ev in evs if ev.time == a)


def select_events_between(evs: Dict[str, Event], a, b) -> Iterator[Event]:
    if a is ... and b is ...:
        return iter(evs.values())
    elif b is ...:
        return (ev for k, ev in evs if a <= ev.time)
    elif a is ...:
        return (ev for k, ev in evs if ev.time <= b)
    else:
        return (ev for k, ev in evs if a <= ev.time <= b)



from pydemic.testing import en


def pytest_configure(config):
    config.addinivalue_line("markers", "slow: mark tests that take long to run")
    config.addinivalue_line(
        "markers", "external: mark tests that use external resources from the internet"
    )



import numpy as np
import pandas as pd

from .smoothing import smoothed_diff
from ..formulas import formula, get_function

FORMULA_OPTS = {
    "positional": 1,
    "options": ("dt",),
    "invalid": "pass",
    "formula_name": "epidemic_curve",
}


@formula(positional=1, invalid="pass", options=("smooth",))
def infectious_curve(cases: pd.DataFrame, gamma=None, smooth=True, **kwargs):
    """
    Infer the infectious curve from the curve of cumulative cases.

    Return the "(I)nfectious" compartment from case data by propagating the
    number of new cases to the whole duration of its infectious period.

    This transformation is necessary because neither the cumulative cases nor the
    number of new cases per day correspond directly to any compartment in the
    traditional Differential equation models.

    Args:
        cases:
            Curve with the accumulative number of cases.
        gamma:
            Inverse of infectious period. Can be a number or an object that
            return parameters using the :func:`pydemic.params.get_param`
            function.
        smooth:
            If False, skip smoothing step.
    """
    cases = np.asarray(cases)
    N = len(cases)
    infectious = np.zeros(N)

    if smooth:
        new_cases = smoothed_diff(cases, **kwargs)
    else:
        new_cases = np.diff(cases, prepend=0.0)

    for i, x in enumerate(new_cases):
        infectious[i:] += x * np.exp(-gamma * np.arange(0, N - i))

    return infectious


def epidemic_curve(model, cases, *args, **kwargs):
    """
    Return the epidemic curve from cases for the given model.

    Args:
        model:
            A string like "SIR", "SEIR" or "SEAIR" identifying model type.
        cases:
            A dataframe of "cases" and "deaths" columns with cumulative values..

    Keyword Args:
        Keyword arguments are passed to the work function that implements
        each model.

    Returns:
        A dataframe that can be used to initialize model.

    See Also:
        :func:`sir_curves`
        :func:`seir_curves`
        :func:`seair_curves`
    """
    fn = get_function(model, "epidemic_curve")
    return fn(cases, *args, **kwargs)


@formula("SIR", **FORMULA_OPTS)
def sir_curves(cases, gamma, population, dt=1, **kwargs):
    """
    Infer SIR curves from experimental data.
    """

    N = len(cases)
    infectious = infectious_curve(cases, gamma=gamma, **kwargs)

    # Allocate data and initialize with values
    data = np.zeros((N, 3), dtype=float)
    iS, iI, iR = range(3)
    data[0, iS] = population
    data[:, iI] = infectious

    for i, row in enumerate(data[:-1], start=1):
        S, I, R = row
        data[i, iR] = R + gamma * I * dt
        data[i, iS] = population - data[i, 1:].sum()

    columns = ("susceptible", "infectious", "recovered")
    return pd.DataFrame(data, columns=columns, index=cases.index)


@formula("SEIR", **FORMULA_OPTS)
def seir_curves(cases, gamma, sigma, population, dt=1, **kwargs):
    """
    Infer SEIR curves from experimental data.
    """

    N = len(cases)
    infectious = infectious_curve(cases, gamma=gamma, **kwargs)

    # Allocate data and initialize with values
    data = np.zeros((N, 4), dtype=float)
    iS, iE, iI, iR = range(4)
    data[0, iS] = population
    data[:, iE] = infectious * (sigma / gamma)
    data[:, iI] = infectious

    for i, row in enumerate(data[:-1], start=1):
        S, E, I, R = row
        data[i, iR] = R + gamma * I * dt
        data[i, iS] = population - data[i, 1:].sum()

    columns = ("susceptible", "exposed", "infectious", "recovered")
    return pd.DataFrame(data, columns=columns, index=cases.index)


@formula("SEAIR", **FORMULA_OPTS)
def seair_curves(cases, gamma, sigma, prob_symptoms, rho, population, dt=1, **kwargs):
    """
    Infer SEAIR curves from experimental data.
    """

    N = len(cases)
    infectious = infectious_curve(cases, gamma=gamma, **kwargs)

    # Allocate data and initialize with values
    data = np.zeros((N, 5), dtype=float)
    iS, iE, iA, iI, iR = range(5)
    data[0, iS] = population
    data[:, iE] = infectious * (sigma / gamma / prob_symptoms)
    data[:, iA] = infectious * (1 - prob_symptoms) / prob_symptoms
    data[:, iI] = infectious

    for i, row in enumerate(data[:-1], start=1):
        S, E, A, I, R = row
        data[i, iR] = R + gamma * (I + A) * dt
        data[i, iS] = population - data[i, 1:].sum()

    columns = ("susceptible", "exposed", "asymptomatic", "infectious", "recovered")
    return pd.DataFrame(data, columns=columns, index=cases.index)



from .disease_class import Disease
from .utils import lazy_stored_string


class Covid19(Disease):
    """
    COVID-19 first cases was detected in 2019 in China and started to spread
    worldwide early 2020. It was shortly afterwards declared a pandemic and
    a global health crisis.
    """

    name = "COVID-19"
    full_name = "SARCov-19"

    MORTALITY_TABLE_DEFAULT = "verity"
    HOSPITALIZATION_TABLE_DEFAULT = "verity"
    MORTALITY_TABLE_DESCRIPTIONS = {
        "verity": lazy_stored_string("covid-19/mortality-table-verity.txt")
    }

    def _epidemic_curve(self, region, api="auto", **kwargs):
        """
        Load epidemic curve for the given region from the internet.
        """
        from ..api.covid19 import epidemic_curve

        return epidemic_curve(region, api, **kwargs)

    def R0(self, **kwargs):
        return 2.74

    def rho(self, **kwargs):
        return 0.45

    def icu_period(self, **kwargs):
        return 7.5

    def hospitalization_period(self, **kwargs):
        return 10.0

    def critical_delay(self, **kwargs):
        # FIXME: find reference.
        # We estimated this parameter from anecdotal evidence from physicians
        # that work in hospitals and non-official compilations of cases in
        # Brazil.
        #
        # This parameter should be available in the literature, however.
        return 1.0

    def severe_delay(self, **kwargs):
        return 3.3

    def hospitalization_overflow_bias(self, **kwargs):
        # We do not have this parameter and it seems very unlikely that it will
        # ever be measured directly. This correspond to the fraction of
        # hospitalized cases that evolve to critical state in the absence of
        # treatment *in excess* of the natural evolution of the disease.
        return 0.25

    def infectious_period(self, **kwargs):
        return 3.47

    def incubation_period(self, **kwargs):
        return 3.69



import numpy as np

from . import sir
from .base import formula

IGNORE = tuple(x for x in sir.IGNORE if x != "sigma")
seir_formula = lambda ignore=(): formula("SEIR", ignore=(*ignore, *IGNORE))

R0 = sir.R0.register("SEIR")
beta = sir.beta.register("SEIR")


@seir_formula()
def R0_from_K(K, gamma, sigma) -> float:
    """
    Return R0 from the exponential growth factor K and the other model
    parameters.
    """
    return 1.0 + (gamma + sigma + K) * K / (gamma * sigma)


@seir_formula()
def K(R0, gamma, sigma) -> float:
    """
    Return the exponential growth factor K from R0 and the other model
    parameters.
    """
    mu = sigma + gamma
    return 0.5 * mu * (np.sqrt(1 + 4 * (R0 - 1) * sigma * gamma / (mu * mu)) - 1)


#
# Initial conditions
#
@seir_formula()
def time_to_seed(cases, R0, gamma, sigma):
    """
    Simple exponential extrapolation for the time from the first seed to
    reaching the given number of cases using the SEIR/SEAIR model.

    This is a decent approximation for the outset of an epidemic. If the seed
    consists of more than a single individual, divide the number of cases by
    the estimated seed size.
    """
    if R0 <= 1:
        raise ValueError(f"R0 must be greater than one (got {R0})")

    seed = 1
    cases_ = (cases - seed) / seed
    k = K.formula(R0, gamma, sigma)
    beta = gamma * R0
    return 1 / k * np.log((cases_ * k + beta) / beta)


@seir_formula()
def infectious_from_cases(cases, R0, gamma, sigma):
    """
    Initializes the "infectious" component of a SEIR/SEAIR model from the current
    number of cases.

    This formula assumes a perfect exponential growth.
    """
    if R0 <= 1:
        raise ValueError(f"R0 must be greater than one (got {R0})")

    seed = 1
    k = K.formula(R0, gamma, sigma)
    beta = gamma * R0
    return ((cases - seed) * k + beta * seed) / beta


@seir_formula()
def exposed_from_cases(cases, R0, gamma, sigma):
    """
    Initializes the "exposed" component of a SEIR model from the current
    number of cases.

    This formula assumes a perfect exponential growth.
    """
    k = K.formula(R0, gamma, sigma)
    I0 = infectious_from_cases.formula(cases, R0, gamma, sigma)
    return I0 * (gamma * R0) / (k + sigma)


@seir_formula()
def state_from_cases(population, cases, R0, gamma, sigma):
    """
    Initialize the 4 components of a SEIR model from cases and population.
    """
    I = infectious_from_cases.formula(cases, R0, gamma, sigma)
    E = exposed_from_cases.formula(cases, R0, gamma, sigma)
    R = cases - I
    S = population - I - E - R
    return np.array([S, E, I, R])



import numpy as np
from ..packages import click

from .abstract_seir import AbstractSEIR
from .ode_model import ODEModel


class SEIR(ODEModel, AbstractSEIR):
    """
    A simple SIR model linearized around the DFE.
    """

    def diff(self, x, t):
        s, e, i, r = x
        n = s + e + i + r
        beta = self.beta
        gamma = self.gamma
        sigma = self.sigma

        return np.array(
            [
                -beta * s * (i / n),
                +beta * s * (i / n) - sigma * e,
                +sigma * e - gamma * i,
                +gamma * i,
            ]
        )


if __name__ == "__main__":

    @click.command()
    @click.option("--R0", "-r", default=2.74, help="Reproduction number")
    @click.option("--duration", "-t", default=90, type=int, help="Duration")
    @click.option("--log", default=False, type=bool, help="Use log-scale?")
    def cli(duration, r0, log):
        m = SEIR()
        m.R0 = r0
        m.run(duration)
        m.plot(logy=log, show=True)

    cli()



import numpy as np
import pandas as pd


def delayed(data, delay, K=0):
    """
    Transform data to be delayed by the given time delay.
    """
    ts = data.index

    if K and data.iloc[0]:
        ts_, data = extend_data(data, ts, delay, max(K, 0.0))
    else:
        ts_ = ts + delay

    return pd.Series(np.interp(ts, ts_, data), index=ts)


def delayed_with_discharge(data, dt1, dt2, K=0, positive=False):
    """
    Similar to :func:`delayed`, but includes a discharge time.

    Times dt1 and dt2 are additive, that is, elements wait a period dt1 to
    enter the delayed state, then wait dt2 to leave it. This function returns
    the difference between these two states.
    """
    ts = data.index

    if dt1 == 0:
        entry = data
    else:
        entry = np.interp(ts, *extend_data(data, ts, dt1, K))

    if dt2 == 0:
        discharge = entry
    else:
        discharge = np.interp(ts, *extend_data(entry, ts, dt2, K))

    diff = entry - discharge
    return pd.Series(np.maximum(diff, 0) if positive else diff, index=ts)


def extend_data(data, ts, delay, K):
    if K == "infer":
        K = np.log(data.iloc[1] / data.iloc[0])

    ts_pre = ts - delay
    ts_pre = ts_pre[ts_pre < 0]
    n = len(ts_pre)
    ts_ = np.concatenate([ts_pre + n, ts + delay])

    data_pre = data.iloc[0] * np.exp(K * ts_[:n] - K * ts_[n])
    data = np.concatenate([data_pre, data])
    return ts_, data



import mundi
from pydemic.diseases import covid19
from pydemic.models import SEIR


class TestClinicalModel:
    def test_clinical_model_uses_region_IFR(self, region="BR"):
        m = SEIR(region=region, disease=covid19)
        br = mundi.region("BR")
        m1 = m.clinical.crude_model()
        m2 = m.clinical.delay_model()
        m3 = m.clinical.overflow_model()

        IFR = covid19.IFR(region=region)
        CFR = covid19.CFR(region=region)

        for m in [m1, m2, m3]:
            assert m.region == br
            assert m.IFR == m.infection_fatality_ratio == IFR
            assert m.CFR == m.case_fatality_ratio == CFR



from pprint import pformat


def main(cls, *args, **kwargs):
    """
    Executes the default action for the model. Convenient for making quick
    and dirt CLI tools.
    """
    import click

    kind_map = {"int": int, "float": float, "str": str}

    @click.option("--plot", is_flag=True, help="Display plot")
    @click.option("--debug", is_flag=True, help="Display debug information")
    def cli(plot=False, debug=False, **kwargs_):
        kwargs_ = {k: v for k, v in kwargs_.items() if v is not None}
        kwargs_ = {**kwargs, **kwargs_}
        m = cls._main(*args, **kwargs_)
        m.run()
        print(m)
        if debug:
            print("\n\nDEBUG SYMBOLS")
            for k, v in vars(m).items():
                print(k, "=", pformat(v))
        if plot:
            m.plot(show=True)

    for cmd, help in list(cls.OPTIONS.items())[::-1]:
        cmd, _, kind = cmd.partition(":")
        kind = kind_map[kind or "float"]
        cmd = cmd.replace("_", "-")
        cli = click.option(f"--{cmd}", help=help, type=kind)(cli)
    cli = click.command()(cli)
    cli()


def _main(cls, *args, **kwargs):
    return cls(*args, **kwargs)



from typing import Sequence

import numpy as np
import pandas as pd
import statsmodels.api as sm
from statsmodels.regression.rolling import RollingOLS

from mundi.pandas import Pandas
from .utils import diff, smooth, cases
from ..docs import docstring
from ..types import ValueStd

KT_METHODS = {}
K_METHODS = {}
METHODS = {"K": K_METHODS, "Kt": KT_METHODS}
ARGS = """Args:
    curves:
        Cumulative series of cases or deaths.
    window:
        Window size of triangular smoothing. Can be a single number or
        a 2-tuple with the window used to compute daily cases and the window
        used to smooth out the log-derivative.
"""


def method(which, names):
    """
    Decorator that register function as a Kt or K inference method.
    """

    names = [names] if isinstance(names, str) else names
    db = METHODS[which]

    def decorator(fn):
        for name in names:
            if name in db:
                raise RuntimeError(f"method already exists: {name}")
            db[name] = fn
        return fn

    return decorator


#
# Main interface
#
@docstring(args=ARGS)
def estimate_Kt(curves: Pandas, window=14, method="RollingOLS", **kwargs) -> pd.DataFrame:
    """
    Compute K(t) from the epidemic curves. This is function is just a façade to
    the concrete implementations listed bellow.

    {args}
        method ('ROLS', 'naive'):
            The method name:
            * 'ROLS'/'RollingOLS': Uses a Rolling Ordinary Least Squares
              implementation from :func:`rolling_OLS_Kt`.
            * 'naive': Obtain K(t) from the derivative of a smoothed out curve
               of log-new-cases. Implemented by :func:`naive_Kt`.

    Returns:
        pd.DataFrame with a Kt column and possibly other information (e.g.,
        Kt_low, Kt_high) depending on the method.

    See Also:
        naive_Kt
        rolling_OLS_Kt
    """
    return KT_METHODS[method](curves, window=window, **kwargs)


@docstring(args=ARGS)
def estimate_K(curves: Pandas, method="naive", **kwargs) -> ValueStd:
    """
    Compute an stationary K for the entire epidemic curve. This is function is
    just a façade to the concrete implementations listed bellow.

    {args}
        method ('OLS', 'naive'):
            The method name:
            * 'OLS': Uses ordinary least squares of log-new-cases.
            * 'Naive': Simply compute the mean and standard deviation from
              the smoothed-out curve or log-new-cases.

    See Also:
        naive_K
        OLS_K
    """
    return K_METHODS[method](curves, **kwargs)


#
# Time-dependent K(t)
#
@docstring(args=ARGS)
@method("Kt", "naive")
def naive_Kt(curves: Pandas, window=14) -> pd.DataFrame:
    """
    Return K(t) as the derivative of the logarithm of a smoothed-out series of
    new cases or deaths.

    {args}
    """
    a, b = window if isinstance(window, Sequence) else (window, window)
    daily = diff(cases(curves), prepend=0, smooth=a)
    data = diff(np.log(daily), smooth=b)

    return pd.DataFrame({"Kt": data})


@docstring(args=ARGS)
@method("Kt", ("ROLS", "RollingOLS"))
def rolling_OLS_Kt(curves, window=14) -> pd.DataFrame:
    """
    A Rolling window Ordinary Least Squares inference of the derivative of the
    logarithm of the number of cases.

    {args}
    """

    a, b = window if isinstance(window, Sequence) else (window, window)
    daily = diff(cases(curves), smooth=a)

    # We first make a OLS inference to extrapolate series to past
    Y = np.log(daily).values
    X = np.arange(len(Y))
    ols = sm.OLS(Y[:b], sm.add_constant(X[:b]), missing="drop")
    res = ols.fit()

    # We need at least c new observations to obtain a result without NaNs
    m = res.params[1]

    X_ = np.arange(X[0] - b, X[0])
    Y_ = m * (X_ - X[0]) + Y[0]

    X = np.concatenate([X_, X])
    Y = np.concatenate([Y_, Y])

    # Use Rolling OLS to obtain an inference to the growth ratio
    ols = RollingOLS(Y, sm.add_constant(X), window=b, missing="drop")
    res = ols.fit()

    Kt = res.params[b:, 1]
    low, high = res.conf_int()[b:, :, 1].T

    out = pd.DataFrame({"Kt": Kt, "Kt_low": low, "Kt_high": high}, index=curves.index)

    return out


#
# Static K
#
@docstring(args=ARGS)
@method("K", "naive")
def naive_K(curves, window=14) -> ValueStd:
    """
    Ordinary Least squares inference of data growth factor using the logarithm
    of the new number of cases.

    {args}
    """
    data = naive_Kt(curves, window)["Kt"]
    return ValueStd(data.mean(), data.std())


@docstring(args=ARGS)
@method("K", "OLS")
def ordinary_least_squares_K(curves, window=14) -> ValueStd:
    """
    Ordinary Least squares inference of data growth factor using the logarithm
    of the new number of cases.

    {args}
    """

    a, b = window if isinstance(window, Sequence) else (window, window)
    daily = diff(cases(curves), a)

    Y = smooth(np.log(daily), b)
    X = np.arange(len(Y))

    ols = sm.OLS(Y, sm.add_constant(X), missing="drop")
    res = ols.fit()

    return ValueStd(res.params[1], res.cov_HC0[1, 1])



from numbers import Number, Real
from types import SimpleNamespace
from typing import NamedTuple, Optional, Callable, Union, Any

import sidekick as sk

ParamLike = Any
NOT_GIVEN = object()


class Param(NamedTuple):
    """
    Represents a parameter
    """

    data: Number
    ref: Optional[str] = None
    pdf: Optional[Union[Callable, str]] = None

    @property
    def value(self):
        v = self.data
        return v() if callable(v) else v

    def __str__(self):
        suffix = []
        if self.ref:
            suffix.append(str(self.ref))
        if self.pdf and not isinstance(self.pdf, SimpleNamespace):
            suffix.append(str(self.pdf))
        suffix = ", ".join(suffix)
        return f"{self.value} ({suffix})" if suffix else str(self.value)


class ParamMeta(type):
    @sk.lazy
    def getters(cls):
        getters = {}
        for k, v in vars(cls).items():
            if hasattr(v, "fget"):
                getters[k] = v.fget
        return getters


class Params(metaclass=ParamMeta):
    """
    Represents a set of parameters.
    """

    __slots__ = ("name", "refs", "pdfs", "__dict__")

    def __init__(self, name=None, **kwargs):
        type(self).name.__set__(self, name or type(self).__name__)
        type(self).refs.__set__(self, {})
        type(self).pdfs.__set__(self, {})

        for k, v in kwargs.items():
            ref = rvs = None
            if isinstance(v, tuple):
                v, ref, rvs = v
            super().__setattr__(k, v)
            if ref is not None:
                self.refs[k] = ref
            if rvs is None:
                rvs = SimpleNamespace(rvs=cte(v))
            self.pdfs[k] = rvs

    def __setattr__(self, k, v):
        raise TypeError(
            f"Parameters are immutable. Use param.copy({k}={v!r}) to create a copy with "
            f"different values"
        )

    def __getitem__(self, item):
        try:
            return self.__dict__[item]
        except KeyError:
            pass
        prop = type(self).getters[item]
        return prop(self)

    def __iter__(self):
        refs = self.refs.get
        rvs = self.pdfs.get
        for k, v in self.__dict__.items():
            yield k, Param(v, refs(k), rvs(k))

    def __str__(self):
        return self.summary()

    def __repr__(self):
        args = (f"{k}={v.value!r}" for k, v in self)
        args = ", ".join(args)
        cls = type(self).__name__
        return f"{cls}({self.name!r}, {args})"

    def copy(self, **kwargs):
        """
        Copy, possibly overwriting some values.
        """
        cls = type(self)
        for k, v in kwargs.items():
            if hasattr(self, k):
                kwargs[k] = v
            else:
                raise AttributeError(f"invalid attribute: {k}")

        kwargs = {**self.__dict__, **kwargs}
        return cls(self.name, **kwargs)

    def value(self, key, default: Number = NOT_GIVEN) -> Number:
        """
        Return the value of the given parameter.
        """
        try:
            value = getattr(self, key)
            return value() if callable(value) else value
        except KeyError:
            if default is NOT_GIVEN:
                raise ValueError(f"Invalid parameter: {key}")
            return default

    def ref(self, key) -> Optional[str]:
        """
        Return the reference for the given parameter, if it exists.
        """
        return self.refs.get(key)

    def pdf(self, key):
        """
        Return the probability distribution function for the given param.
        """
        return self.pdfs.get(key)

    def param(self, key) -> Param:
        """
        Return a :cls:`Param` object describing the given parameter.
        """
        value = self.value(key)
        ref = self.ref(key)
        pdf = self.pdf(key)
        return Param(value, ref, pdf)

    def summary(self):
        """
        Print a summary of all parameters in the set.
        """
        lines = [f"Parameters ({self.name}):", *(f"  - {k}: {p}" for k, p in self)]
        return "\n".join(lines)


# Return a function always return the same value
cte = lambda v: lambda: v


def param(value: ParamLike, ref=None, pdf=None) -> Param:
    """
    Declares a parameter with optional reference attribution and RVS
    attribution.
    """
    if isinstance(value, Param):
        ref = ref or value.ref
        pdf = pdf or value.pdf
        value = value.value

    return Param(value, ref, pdf)


def get_param(name: str, params: ParamLike, default: Real = None) -> Real:
    """
    Get a parameter from a params object.

    Args:
        name:
            Parameter name.
        params:
            A params object. It can be Python object that stores parameters as
            attributes, methods or keys.
        default:
            Default value if parameter is not found in the params object.

    Returns:
        A parameter value.

    See Also:
        :func:`select_param`
    """
    try:
        value = getattr(params, name)
        return value() if callable(value) else value
    except AttributeError:
        pass

    try:
        return params[name]
    except (KeyError, TypeError):
        if default is None:
            raise ValueError(f"Parameter '{name}' not found and no default was given")
        return default


def select_param(name: str, params: ParamLike, value: Number = None) -> Real:
    """
    Similar to :func:`get_param`, but the explicitly passed value takes precedence.

    Args:
        name:
            Parameter name.
        params:
            A params object like in the :func:`get_param` function.
        value:
            An optional value override. If value is None, fetches value from
            the params object.

    Returns:
        A parameter value.

    See Also:
        :func:`get_param`
    """
    return get_param(name, params) if value is None else value



from abc import ABC

from .abstract_seir import AbstractSEIR as Base
from ..packages import sk, np
from ..utils import state_property, param_property


class AbstractSEAIR(Base, ABC):
    """
    Abstract base class for all SEIR-based models.
    """

    class Meta:
        model_name = "SEAIR"
        data_aliases = {"A": "asymptomatic"}

    # Basic epidemiological parameters
    rho = param_property("rho", default=1.0)
    prob_symptoms = param_property("prob_symptoms", default=1.0)
    Qs = sk.alias("prob_symptoms", mutable=True)

    # Derived expressions
    @property
    def beta(self):
        gamma = self.gamma
        R0 = self.R0
        rho = self.rho
        Qs = self.Qs
        return gamma * R0 / (Qs + (1 - Qs) * rho)

    # Simulation state
    susceptible = state_property(0)
    exposed = state_property(1)
    asymptomatic = state_property(2)
    infectious = state_property(3)
    recovered = state_property(4)

    def _initial_state(self):
        return np.array((self.population - 1, 1, 0, 0, 0), dtype=float)

    def get_data_force(self, idx):
        I = self["infectious", idx]
        A = self["asymptomatic", idx]
        N = self["N", idx]
        return self.beta * (I + self.rho * A) / N



import weakref
from abc import ABC, ABCMeta
from typing import Any

from ..logging import log


class PropertyMeta(ABCMeta):
    """
    Metaclass for property objects.

    Makes class behaves like a descriptor.
    """

    def __get__(cls, instance, kind=None):
        if instance is None:
            return cls
        else:
            return cls(instance)

    def __set_name__(cls, owner, name):
        cls.__dict__.setdefault("name", name)

    def as_property(cls):
        """
        Explicitly asks for a property interface in class declaration.

        >>> class Foo:
        ...     prop: Property = Property.as_property()
        """
        return property(cls)

    def patch(cls, klass: type = None, verbose=True):
        """
        Patch class with all public and private members of class.

        Dunder methods and attributes are omitted.

        Examples:
            >>> @Property.patch()
            ... class Mixin:
            ...     def method1(self):
            ...         pass
            ...     def method2(self):
            ...         pass
        """
        if klass is None:
            return lambda kind: cls.patch(kind)

        methods = set()
        for k, v in klass.__dict__.items():
            cls.add_method(v, k, verbose=False)
            methods.add(k)

        if verbose:
            cls_name = cls.__name__
            methods = ", ".join(map(repr, methods))
            log.info(f"{cls_name} patched with the {methods} methods")

        return klass

    def add_method(cls, func=None, name=None, verbose=True):
        """
        Patch class to include the given function as method. This is useful
        under interactive workloads until code is moved back to upstream.

        Examples:
            >>> @Property.add_method()
            ... def new_method(self, *args, **kwargs):
            ...     print("Property now as a .new_method method.")
        """
        if func is None:
            return lambda fn: cls.add_method(fn, name)

        if name is None:
            name = func.__name__
        setattr(cls, name, func)

        if verbose:
            cls_name = cls.__name__
            log.info(f"{cls_name} patched with a '{name}' method")

        return func


class BaseProperty(ABC, metaclass=PropertyMeta):
    """
    Base class for the several property accessors used in the Pydemic API.
    """

    __slots__ = ()
    _object: Any
    name: str

    def __repr__(self):
        try:
            name = self.name
        except AttributeError:
            name = type(self).__name__.lower()
            if name.endswith("property"):
                name = name[:8]

        cls = type(self._object).__name__
        return f"<{cls}.{name} property>"

    def __getstate__(self):
        return self._object

    def __setstate__(self, state):
        # noinspection PyArgumentList
        self.__init__(state)

    def __eq__(self, other):
        if type(self) is type(other):
            other: BaseProperty
            same_name = getattr(self, "name", None) == getattr(other, "name", None)
            return same_name and self._object == other._object
        return NotImplemented


class Property(BaseProperty):
    """
    Default base for pydemic properties namespaces.

    Uses weakrefs to avoid circular memory dependencies.
    """

    __slots__ = ("_ref",)
    _object = property(lambda self: self._ref())

    def __init__(self, obj):
        self._ref = weakref.ref(obj)


class StrongProperty(BaseProperty):
    """
    A version of Property for objects that do not support weak references.
    """

    __slots__ = ("_object",)

    def __init__(self, obj):
        self._object = obj



from types import SimpleNamespace

import sidekick as sk
from sidekick import placeholder as _

from .param import Params


class EpidemicParams(Params):
    R0: float
    infectious_period: float
    incubation_period: float = 0.0
    gamma = sk.property(1 / _.infectious_period)
    sigma = sk.property(1 / _.incubation_period)


#
# Epidemiological parameters
#
EPIDEMIC_DEFAULT = ep = EpidemicParams(
    "Default", R0=2.74, rho=0.55, prob_symptoms=0.14, incubation_period=3.69, infectious_period=3.47
)

epidemic = SimpleNamespace(
    DEFAULT=EPIDEMIC_DEFAULT,
    DEFAULT_SIR=EpidemicParams(
        "SIR Default", R0=ep.R0, infectious_period=ep.incubation_period + ep.infectious_period
    ),
)



import locale
from datetime import timedelta

import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_array_equal
from pandas.testing import assert_frame_equal

from pydemic import utils

DAY = timedelta(days=1)


class TestUtilityFunctions:
    def test_format_functions_en_US(self):
        fmt = utils.fmt

        try:
            locale.setlocale(locale.LC_ALL, "en_US.UTF-8")
        except locale.Error:
            return pytest.skip()

        assert fmt(None) == "-"
        assert fmt(float("inf")) == "infinity"
        assert fmt(0.10) == "0.1"
        assert fmt(0.12) == "0.12"
        assert fmt(0.01) == "0.01"
        assert fmt(0.012) == "0.012"
        assert fmt(0.0123) == "0.012"
        assert fmt(0.00123) == "1.23e-03"
        assert fmt(0.0012) == "1.2e-03"
        assert fmt(1.2341) == "1.23"
        assert fmt(12.341) == "12.34"
        assert fmt(123.41) == "123.4"
        assert fmt(1234) == "1,234"
        assert fmt(1234.5) == "1,234"
        assert fmt(42_123.1) == "42,123"
        assert fmt(42_123) == "42,123"
        assert fmt(1_000_000) == "1M"
        assert fmt(10_000_000) == "10M"
        assert fmt(12_000_000) == "12M"
        assert fmt(12_300_000) == "12.3M"
        assert fmt(12_340_000) == "12.34M"
        assert fmt(12_341_000) == "12.34M"
        assert fmt(-12_341_000) == "-12.34M"
        assert fmt(123_456_000) == "123.5M"
        assert fmt(1_234_567_000) == "1.23B"

    def test_format_functions_pt_BR(self):
        fmt = utils.fmt
        try:
            locale.setlocale(locale.LC_ALL, "pt_BR.UTF-8")
        except locale.Error:
            return pytest.skip()

        assert fmt(None) == "-"
        assert fmt(0.10) == "0,1"
        assert fmt(0.12) == "0,12"
        assert fmt(0.01) == "0,01"
        assert fmt(0.012) == "0,012"
        assert fmt(0.0123) == "0,012"
        assert fmt(0.00123) == "1,23e-03"
        assert fmt(0.0012) == "1,2e-03"
        assert fmt(1.2341) == "1,23"
        assert fmt(12.341) == "12,34"
        assert fmt(123.41) == "123,4"
        assert fmt(1234) == "1.234"
        assert fmt(1234.5) == "1.234"
        assert fmt(42_123.1) == "42.123"
        assert fmt(42_123) == "42.123"
        assert fmt(1_000_000) == "1M"
        assert fmt(10_000_000) == "10M"
        assert fmt(12_000_000) == "12M"
        assert fmt(12_300_000) == "12,3M"
        assert fmt(12_340_000) == "12,34M"
        assert fmt(12_341_000) == "12,34M"
        assert fmt(-12_341_000) == "-12,34M"
        assert fmt(123_456_000) == "123,5M"
        assert fmt(1_234_567_000) == "1,23B"

    def test_other_formats(self):
        assert utils.pc(0.5) == "50%"
        assert utils.pm(0.05) == "50‰"
        assert utils.p10k(0.005) == "50‱"
        assert utils.p100k(0.0005) == "50/100k"
        assert utils.safe_int(3.14) == 3
        assert utils.safe_int(float("nan")) == 0

    def test_text_functions(self):
        assert utils.indent("foo\nbar", 2) == "  foo\n  bar"
        assert utils.slugify("Foo Bar") == "foo-bar"

    def test_format_args(self):
        assert utils.format_args(1, 2, op="sum") == "1, 2, op='sum'"


class TestDatetime:
    def test_now_and_today(self):
        assert utils.now().date() == utils.today()


class TestSequenceFunctions:
    def test_flatten_and_unflatten_dict(self):
        flat = {"foo.bar.bat": 2, "foo.bar.baz": 1, "foo.bot": 3}
        nested = {"foo": {"bar": {"baz": 1, "bat": 2}, "bot": 3}}
        assert utils.flatten_dict(nested) == flat
        assert utils.unflatten_dict(flat) == nested


class TestJson:
    def test_convert_to_json(self):
        for x in [True, None, "string", 42, 3.14, ["list", "of items"], {"dict": "too"}]:
            assert utils.to_json(x) == x

        assert utils.to_json({0: 0, 3.14: 1, True: 2, None: 3, ...: 4}) == {
            "0": 0,
            "3.14": 1,
            "true": 2,
            "null": 3,
            "...": 4,
        }

    def test_to_json_as_method(self):
        class Foo:
            def to_json(self):
                return {"type": "Foo"}

        foo = Foo()
        assert utils.to_json(foo) == {"type": "Foo"}

    def test_to_json_raises_a_type_error_in_unsupported_types(self):
        class Foo:
            pass

        foo = Foo()
        with pytest.raises(TypeError):
            utils.to_json(foo)

        with pytest.raises(TypeError):
            utils.to_json({(1, 2): 3})


class TestDataFrame:
    def test_trim_zeros(self):
        data1 = [0, 0, 0, 1, 2, 0, 3, 0, 0]
        data2 = pd.Series(data1)
        data3 = [1, *data1, 2]
        assert_array_equal(utils.trim_zeros(data1), [1, 2, 0, 3])
        assert_array_equal(utils.trim_zeros(data2), [1, 2, 0, 3])
        assert_array_equal(utils.trim_zeros(data3), data3)
        assert_array_equal(utils.trim_zeros([0, 0, 0]), [])


class TestFunctions:
    def test_maybe_run(self):
        from math import sqrt

        assert utils.maybe_run(sqrt, 4) == 2
        assert utils.maybe_run(sqrt, None) is None

    def test_coalesce(self):
        with pytest.raises(ValueError):
            utils.coalesce(None, None, raises=True)

        with pytest.raises(ValueError):
            utils.coalesce(raises=True)

        assert utils.coalesce() is None
        assert utils.coalesce(1, 2) == 1
        assert utils.coalesce(1, None) == 1
        assert utils.coalesce(None, 1) == 1

    def test_interpolant(self):
        x = [1, 2, 3]
        y = [1, 2, 4]
        fn = utils.interpolant(x, y)

        assert fn(1.5) == 1.5
        assert fn(2.5) == 3.0
        assert_array_equal(fn([1.5, 2.5]), [1.5, 3.0])

    def test_lru_safe_cache(self):
        compute = False

        @utils.lru_safe_cache(10)
        def fn(n):
            nonlocal compute
            compute = True
            return [n] * n

        assert fn(2) == [2, 2]
        assert compute is True

        compute = False
        assert fn(2) == [2, 2]
        assert compute is False

        assert fn(2) is not fn(2)


class TestTimeseries:
    def get_index(self, size, weekday=0):
        monday = pd.to_datetime("2020-05-25")
        start = monday + DAY * weekday
        return pd.to_datetime([start + n * DAY for n in range(size)])

    def get_series(self, size, weekday=0):
        index = self.get_index(size, weekday)
        return pd.Series(np.arange(size), index=index)

    def get_dataframe(self, size, weekday=0):
        index = self.get_index(size, weekday)
        return pd.DataFrame(
            {"even": np.arange(0, 2 * size, 2), "odd": np.arange(1, 2 * size, 2)}, index=index
        )

    def test_trim_weeks(self):
        data = self.get_dataframe(18)
        clean = utils.trim_weeks(data)
        assert len(clean) == 14
        assert_frame_equal(clean, data.iloc[:14])

        clean = utils.trim_weeks(data, 1)
        assert len(clean) == 14
        assert_frame_equal(clean, data.iloc[1:15])

    def test_accumulate_weekly(self):
        data = self.get_dataframe(25, 0)
        acc = utils.accumulate_weekly(data, "trim")
        assert_array_equal(data.iloc[0:7].sum(), acc.iloc[0])

        data = self.get_dataframe(25, 2)
        acc = utils.accumulate_weekly(data, "trim")
        assert_array_equal(data.iloc[5:12].sum(), acc.iloc[0])

    def test_day_of_week(self):
        data = 1

    def test_weekday_name(self):
        assert utils.weekday_name(0) == "Monday"
        assert_array_equal(utils.weekday_name([0, 2]), ["Monday", "Wednesday"])



from typing import Sequence


def rpartition(seq, n):
    """
    Partition sequence in groups of n starting from the end of sequence.
    """
    seq = list(seq)
    out = []
    while seq:
        new = []
        for _ in range(n):
            if not seq:
                break
            new.append(seq.pop())
        out.append(new[::-1])
    return out[::-1]


def flatten_dict(dic: dict, prefix="", sep=".") -> dict:
    """
    Flatten a nested dictionary into a flat dictionary with dotted namespace.
    """

    out = {}
    for k, v in dic.items():
        k = f"{prefix}{k}"
        if isinstance(v, dict):
            out.update(flatten_dict(v, f"{k}{sep}"))
        else:
            out[k] = v
    return out


def unflatten_dict(dic: dict, sep=".") -> dict:
    """
    Invert the effect of :func:`flatten_dict`
    """

    items = list(dic.items())
    items.reverse()
    out = {}
    while items:
        k, v = items.pop()
        *keys, last_key = k.split(sep)

        d = out
        for k in keys:
            d = d.setdefault(k, {})
        d[last_key] = v

    return out


def extract_keys(keys, dic, drop=True):
    """
    Extract keys from dictionary and return a dictionary with the extracted
    values.

    If key is not included in the dictionary, it will also be absent from the
    output.
    """

    out = {}
    for k in keys:
        try:
            if drop:
                out[k] = dic.pop(k)
            else:
                out[k] = dic[k]
        except KeyError:
            pass
    return out


def sliced(seq, idx):
    """
    Possibly slice object if index is not None.
    """
    if idx is None:
        return seq
    else:
        try:
            return seq[idx]
        except KeyError:
            raise IndexError(f"invalid index: {idx}")


def is_seq(obj) -> bool:
    """
    Return true if object is a non-string sequence.

    See Also:
        as_seq
    """
    return not isinstance(obj, str) and isinstance(obj, Sequence)


def as_seq(obj) -> Sequence:
    """
    Force object to be a sequence.

    Non-sequence arguments and strings are converted into singleton lists.

    See Also:
        is_seq
    """
    return obj if is_seq(obj) else (obj,)



from typing import Optional, TYPE_CHECKING

import pandas as pd

import mundi
from ..utils import extract_keys

INIT_KEYWORDS = frozenset(["region", "population", "age_distribution", "age_pyramid"])

if TYPE_CHECKING:
    from ..region import RegionT  # noqa: F401


class WithRegionDemography:
    """
    Models with a .region attribute and the possibility of overriding basic
    demographic parameters.
    """

    region: Optional["RegionT"]
    population: float
    age_distribution: Optional[pd.Series]
    age_pyramid: Optional[pd.DataFrame]

    def _init_from_dict(self, kwargs, drop=True):
        """
        Init from dictionary removing used keys.
        """
        opts = extract_keys(INIT_KEYWORDS, kwargs, drop=drop)
        self.__init(**opts)
        return opts

    def __init__(self, **kwargs):
        self.__init(**kwargs)

    def __init(self, region=None, population=None, age_distribution=None, age_pyramid=None):

        if age_distribution is not None and age_pyramid is not None:
            msg = "cannot set age_pyramid and age_distribution simultaneously"
            raise ValueError(msg)

        # Set region
        if region is not None:
            self.region = mundi.region(region)
            population = population or self.region.population
        elif not hasattr(self, "region"):
            self.region = None

        # Set age_pyramid
        if age_pyramid is not None:
            self.age_pyramid = fallback_to_region(age_pyramid, "age_pyramid")
        elif hasattr(self, "age_pyramid"):
            pass
        elif self.region is not None:
            self.age_pyramid = fallback_to_region(self.region, "age_pyramid")
        else:
            self.age_pyramid = None

        # Set age_distribution
        if age_distribution is not None:
            value = fallback_to_region(age_distribution, "age_distribution")
            self.age_distribution = value
        elif hasattr(self, "age_distribution"):
            pass
        elif self.age_pyramid is not None:
            self.age_distribution = self.age_pyramid.sum(1)
            self.age_distribution.name = "age_distribution"
            self.age_distribution.index.names = ["age"]
        elif self.region is not None:
            self.age_distribution = fallback_to_region(self.region, "age_distribution")
        else:
            self.age_distribution = None

        # Set population and fix age_distribution and age_pyramid, if necessary
        if hasattr(self, "population"):
            pass
        elif population is not None:
            population = fallback_to_region(population, "population")
            self.population = population

            if self.age_distribution is not None:
                ratio = population / self.age_distribution.sum()
                if ratio != 1:
                    self.age_distribution *= ratio

            if self.age_pyramid is not None:
                ratio = population / self.age_pyramid.sum().sum()
                if ratio != 1:
                    self.age_pyramid *= ratio
        elif self.age_distribution is not None:
            self.population = self.age_distribution.sum()
        elif self.region is not None:
            self.population = self.region.population
        else:
            self.population = 1_000_000


def fallback_to_region(value, attr):
    """
    Return either value or the given attribute of a region, if value is a
    string or Region instance.
    """
    if value is None:
        return None
    elif isinstance(value, (str, mundi.Region)):
        region = mundi.region(value)
        value = getattr(region, attr)
    try:
        cp = value.copy()
        if value.isna().values.all():
            return None
        return cp
    except AttributeError:
        return value



# flake8: noqa
from .model import ClinicalModel
from .crude_FR import CrudeFR
from .hospitalization_with_delay import HospitalizationWithDelay
from .hospitalization_with_overflow import HospitalizationWithOverflow



import operator
from typing import TYPE_CHECKING, Iterable

import pandas as pd

if TYPE_CHECKING:
    from ..models import Model  # noqa: F401


def map_models(fn: callable, models: Iterable["Model"], key=operator.attrgetter("name")):
    """
    Map function to list of models and return a Series or DataFrame that merges
    the content of fn(model) the given column for each model.

    Args:
        fn:
            Function used to extract data from models.
        models:
            A sequence of models.
        key:
            A callable that receives model and return the corresponding column
            name. If not given, uses `model.name` as key.
    """

    if not isinstance(models, (tuple, list, set)):
        models = tuple(models)
    data = [fn(m) for m in models]
    names = [key(m) for m in models]
    types = set(map(type, data))

    # A sequence of columns should become a dataframe
    if types == {pd.Series}:
        return pd.concat(data, axis=1).set_axis(names, axis=1)

    elif types == {pd.DataFrame}:
        cols = ((name, col) for df, name in zip(data, names) for col in df.columns)
        cols = pd.MultiIndex.from_tuples(cols)
        return pd.concat(data, axis=1).set_axis(cols, axis=1)

    # Otherwise, we assume scalar values and create a simple series object
    return pd.Series(data, index=names)


def prepare_data(models: dict):
    """
    Last step in processing the results of a ModelGroup() getitem.

    Args:
        models:
            A mapping from keys to data.
    """

    data = list(models.values())
    types = set(map(type, data))
    names = list(models.keys())

    # A sequence of columns should become a dataframe
    if types == {pd.Series}:
        return pd.concat(data, axis=1).set_axis(names, axis=1)

    elif types == {pd.DataFrame}:
        cols = ((name, col) for df, name in zip(data, names) for col in df.columns)
        cols = pd.MultiIndex.from_tuples(cols)
        return pd.concat(data, axis=1).set_axis(cols, axis=1)

    # Otherwise, we assume scalar values and create a simple series object
    return pd.Series(data, index=names)


def map_method(*args, **kwargs):
    """
    Given a method name and a sequence of models, return a sequence of values
    of applying the extra positional and keyword arguments in each method.

    Attr:
        attr:
            Method name. Dotted python names are valid.
        models:
            A sequence of models.
        *args, **kwargs:
            Arguments to pass to m.<attr>(*args, **kwargs)
    """
    attr, data, *args = args
    attr, _, method = attr.partition(".")
    if method:
        data = map(operator.attrgetter(attr), data)
        return map(operator.methodcaller(method, *args, **kwargs), data)
    else:
        return map(operator.methodcaller(attr, *args, **kwargs), data)


def model_group_method(attr, out=None):
    """
    A method that applies
    """
    out = out or (lambda x: x)

    def method(self, *args, **kwargs):
        return out(map_method(attr, self.group, *args, **kwargs))

    method.__name__ = attr.rpartition(".")[-1]
    return method


def group(x):
    """
    Convert object to ModelGroup.
    """
    from .model_group import ModelGroup

    return x if isinstance(x, ModelGroup) else ModelGroup(x)



from typing import Optional, FrozenSet, Union, Any

import numpy as np
import pandas as pd
import sidekick as sk

from mundi import Region, region
from .pydemic_property import PydemicProperty
from .pyplot_property import PyplotProperty
from ..diseases import disease as get_disease
from ..properties.decorators import cached

AGGREGATE_METHOD = {
    "population": "sum",
    "age_distribution": "sum",
    "age_pyramid": "sum",
    "icu_capacity": "sum",
    "hospital_capacity": "sum",
    "icu_capacity_public": "sum",
    "hospital_capacity_public": "sum",
}


class CompositePydemicProperty(PydemicProperty):
    region: "CompositeRegion"

    @cached(ttl="epidemic_curve")
    def epidemic_curve(self, disease=None, **kwargs) -> pd.DataFrame:
        fn = get_disease(disease).epidemic_curve
        curves = (fn(r, **kwargs) for r in self.region.regions)
        return self.region.aggregate_values(curves, "sum")


class CompositePyplotProperty(PyplotProperty):
    region: "CompositeRegion"


class CompositeRegion(Region):
    """
    An aggregate of regions.
    """

    name: str
    regions: FrozenSet[Region]
    _keys: FrozenSet[str]

    pydemic = property(CompositePydemicProperty)
    plot = property(CompositePyplotProperty)

    @property
    def id(self):
        try:
            return self.__dict__["id"]
        except KeyError:
            out = "|".join(sorted(map(str, self.regions)))
            self.__dict__["id"] = out
        return out

    def __new__(cls, sub_regions, **kwargs):
        sub_regions = frozenset(map(region, sub_regions))
        if not sub_regions:
            raise ValueError("cannot start with empty list of regions")
        new = object.__new__(cls)
        new.__dict__["regions"] = sub_regions
        new.__dict__["_keys"] = frozenset(kwargs)
        new.__dict__.update(**kwargs)
        return new

    def __hash__(self):
        return hash(self.__getstate__())

    def __getstate__(self):
        args = ((k, self[k]) for k in self._keys)
        return self.regions, args

    def __setstate__(self, state):
        state = self.__dict__
        regions, args = state
        state["regions"] = regions
        for k, v in args:
            state[k] = v

    def __str__(self):
        regions = [r.id for r in self.regions]
        if len(regions) > 5:
            regions = regions[:5] + "..."
        regions = ", ".join(regions)

        return (
            f"Composite Region\n"
            f"id       : {self.id}\n"
            f"name     : {self.name}\n"
            f"type     : {self.type}\n"
            f"subtype  : {self.subtype}\n"
            f"regions  : {regions}"
        )

    def __eq__(self, other):
        if isinstance(other, CompositeRegion):
            keys = self._keys.union(other._keys)
            equal_attrs = (force_bool(self[k] == other[k]) for k in keys)
            return self.regions == other.regions and all(equal_attrs)
        return NotImplemented

    def __repr__(self):
        cls = type(self).__name__
        return f"{cls}({self.regions, !r})"

    def _get_field(self, key):
        method = AGGREGATE_METHOD.get(key, "single")
        return self.aggregate(key, method)

    def aggregate(self, field, by):
        """
        Aggregate values with function.
        """
        values = (region._get_field(field) for region in self.regions)
        return self.aggregate_values(values, by)

    def aggregate_values(self, values, by):
        """
        Given a sequence or iterable with one value per region, aggregate
        values using the given method.
        """
        if by == "sum":
            value, values = sk.uncons(values)
            try:
                value = value.copy()
            except AttributeError:
                return sum(values, value)
            else:
                for v in values:
                    value += v
                return value
        elif by == "sum-pop":
            populations = (region.population for region in self.regions)
            values = (x * pop for x, pop in zip(values, populations))
            return self.aggregate_values(values, "sum")
        elif by == "mean":
            return self.aggregate(values, "sum") / len(self.regions)
        elif by == "mean-pop":
            return self.aggregate(values, "sum-pop") / self.population
        elif by == "single":
            value, *error = set(values)
            if error:
                raise ValueError("regions do not share the same value")
            return value
        else:
            raise ValueError(f"invalid aggregation method: {by}")

    #
    # Hierarchies
    #
    @property
    def parent(self) -> Optional["Region"]:
        """
        Common parent to all regions.
        """
        parent, *error = set(region.parent for region in self.regions)
        if error:
            raise ValueError("regions do not share a common parent.")
        return parent

    def parents(self, dataframe=False):
        if dataframe:
            raise NotImplementedError
        parent = self.parent
        return [parent, *parent.parents()]

    def children(self, dataframe=False, deep=False, which="both"):
        """
        Return list of children.
        """
        if dataframe:
            raise NotImplementedError

        children = {*self.regions}
        if deep:
            for region in self.regions:
                children.update(region.children(deep=True, which=which))
        return list(children)


def force_bool(x: Any) -> bool:
    """
    Similar to bool(x), but call the .all() method until object reduces to a
    real boolean.

    This is useful to test numpy arrays and pandas data structures for equality.
    """
    x: Union[bool, np.ndarray] = bool(x)
    if x is True or x is False:
        return x
    return force_bool(x.all())



from time import sleep

import pytest

from pydemic.cache import ttl_cache


class TestCaches:
    @pytest.mark.slow
    def test_tle_cache_works(self):
        lst = []

        @ttl_cache("testing", timeout=0.01)
        def cache_function_with_side_effects(n):
            print(f"Calling with argument {n}")
            lst.append(n)
            return n

        fn = cache_function_with_side_effects
        fn.clear()

        # Execute w/ side effect
        assert fn(1) == 1
        assert lst == [1]

        # Again...
        assert fn(2) == 2
        assert lst == [1, 2]

        # Now it gets from cache and there is no side-effect
        assert fn(1) == 1
        assert lst == [1, 2]

        # We sleep to invalidate timeout and it should recompute again
        sleep(0.02)
        assert fn(1) == 1
        assert lst == [1, 2, 1]



import datetime
from functools import wraps, lru_cache
from time import time
from typing import Union, Type, Sequence

import sidekick as sk

from pydemic.utils import as_seq
from . import config
from .types import Result
from .utils import today

PERIOD_ALIASES = {
    "day": datetime.timedelta(days=1),
    "week": datetime.timedelta(days=7),
    **{"{n}h": datetime.timedelta(hours=n) for n in range(1, 25)},
}

# TODO: abstract the many available Python caching libs and move this
# functionality to sidekick
#
# References and similar projects
# - http://joblib.readthedocs.io/
# - https://cachetools.readthedocs.io/
# - https://github.com/lonelyenvoy/python-memoization


@sk.fn.curry(2)
def ttl_cache(key, fn, *, timeout=6 * 3600, **cache_kwargs):
    """
    Decorator that creates a cached version of function that stores results
    in disk for the given timeout (in seconds).

    Args:
        timeout:
            Maximum time the item is kept in cache (in seconds).

    Returns:
        A decorated function that stores items in the given cache for the given
        timeout.

    Examples:
        >>> @ttl_cache("my-cache", timeout=3600)
        ... def expensive_function(url):
        ...     # Some expensive function, possibly touching the internet...
        ...     response = requests.get(url)
        ...     ...
        ...     return pd.DataFrame(response.json())

    Notes:
        The each pair of (cache name, function name) must be unique. It cannot
        decorate multiple lambda functions or callable objects with no __name__
        attribute.
    """
    mem = config.memory(key)

    # We need to wrap fn into another decorator to preserve its name and avoid
    # confusion with joblib's cache. This function just wraps the result of fn
    # int a Result() instance with the timestamp as info.
    @mem.cache(**cache_kwargs)
    @wraps(fn)
    def cached(*args, **kwargs):
        return Result(fn(*args, **kwargs), time())

    # Now the decorated function asks for the result in the cache, checks
    # if it is within the given timeout and return or recompute the value
    @wraps_with_cache(fn, cached)
    def decorated(*args, **kwargs):
        mem_item = cached.call_and_shelve(*args, **kwargs)
        result = mem_item.get()
        if result.info + timeout < time():
            mem_item.clear()
            result = cached(*args, **kwargs)
        return result.value

    decorated.clear = mem.clear
    decorated.prune = mem.reduce_size

    return decorated


@sk.fn.curry(2)
def disk_cache(key, fn):
    """
    A simple in-disk cache.

    Can be called as ``simple_cache(key, fn)``, to decorate a function or as as
    decorator in ``@simple_cache(key)``.
    """
    return config.memory(key).cache(fn)


@sk.fn.curry(2)
def period_cache(
    period: Union[str, int, datetime.timedelta],
    fn: callable,
    memory=None,
    maxsize=2048,
    fallback: Sequence[Type[Exception]] = None,
):
    """
    Keeps value in cache within n intervals of the given time delta.

    Args:
        period:
            Time period in which the cache expires. Can be given as a timedelta,
            a integer (in seconds) or a string in the set {'day', 'week', '1h',
            '2h', ..., '24h'}.

            Other named periods cah be registered using the :func:`register_period`
            function.
        fn:
            The decorated function.
        memory (str):
            If given, corresponds to a memory object returned by :func:`memory`.
        maxsize:
            If no memory object is given, corresponds to the maxsize used by
            the lru_cache mechanism.
        fallback:
            If an exception or list of exceptions, correspond to the kinds of
            errors that triggers the cache to check previously stored responses.
            There is nothing that guarantees that the old values will still
            be present, but it gives a second attempt that may hit the cache
            or call the function again.

    Examples:
        >>> @period_cache("day")
        ... def fn(x):
        ...     print('Doing really expensive computation...')
        ...     return ...
    """

    # Select the main method to decorate the cached function
    if memory:
        mem = config.memory(memory)
        decorator = mem.cache
    else:
        decorator = lru_cache(maxsize)

    # Reads a period and return a function that return increments of the period
    # according to the current time. This logic is encapsulated into the key()
    # function.
    date = today()
    ref_time = datetime.datetime(date.year, date.month, date.day).timestamp()
    if isinstance(period, str):
        period = PERIOD_ALIASES[period].seconds
    period = int(period)
    get_time = time
    key = lambda: int(get_time() - ref_time) // period

    # The main cached function. This is stored only internally and the function
    # exposed to the user fixes the _cache_bust and _recur parameters to the
    # correct values.
    fallback = tuple(as_seq(fallback)) if fallback else ImpossibleError

    @decorator
    def cached(_cache_bust, _recur, *args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except fallback:
            if _recur > 0:
                return cached(_cache_bust - 1, _recur - 1, *args, **kwargs)
            raise

    # Save function
    @wraps_with_cache(fn, cached)
    def decorated(*args, **kwargs):
        return cached(key(), 1, *args, **kwargs)

    return decorated


class ImpossibleError(Exception):
    """
    It is an error to raise this exception, do not use it!
    """


def wraps_with_cache(fn, cache=None):
    """
    Like functools.wraps, but also copy the cache methods created either
    by lru_cache or by joblib.Memory.cache.
    """
    cache = cache or fn
    wrapped = wraps(fn)
    for attr in ("cache_info", "clear_cache"):
        if hasattr(cache, attr):
            setattr(wrapped, attr, getattr(cache, attr))
    return wrapped



from pydemic.all import *


# Initialize model
m = SEAIR(region="BR", disease=covid19, R0=1.5)
m.set_ic(cases=20000)
m.run(200)
cm = m.clinical.overflow_model()

# Compute empirical quantities
attack = cm["R:final"] / cm.population
attack_c = cm["cases:final"] / cm.population

print(f"Attack          : {pc(attack)}")
print(f"Attack (c)      : {pc(attack_c)}")
print(f"Qs              : {pc(cm.Qs)}")
print(f"CFR             : {pc(cm.CFR)}")
print(f"CFR (empirical) : {pc(cm.empirical_CFR)}")
print(f"IFR             : {pc(cm.IFR)}")
print(f"IFR             : {pc(cm.infection_fatality_ratio)}")
print(f"IFR (empirical) : {pc(cm.empirical_IFR)}")


# Show plots
plt.subplot(221)
m.plot(log=True)

plt.subplot(222)
cm.plot(["death_rate", "severe"], logy=True)

plt.subplot(223)
cm.plot(["infectious", "cases", "infected", "severe_cases", "critical_cases"], logy=True)

plt.subplot(224)
df = cm[["empirical_CFR", "empirical_IFR"]] * 100
df.plot(ax=plt.gca(), grid=True)

plt.show()



from typing import Sequence
import statsmodels.api as sm
import numpy as np
import pandas as pd

from ..types import ValueStd
from ..utils import trim_zeros
from .. import formulas


def growth_factor(ys: Sequence) -> ValueStd:
    """
    Infer the  growth factor from series of data points ys.

    The growth factor "r" is the parameter that best adjust the sequence:

        y[i + 1] = r * y[i].

    Args:
        ys:
            A sequence of data points approximately generated by the growth
            process shown in the description above.

    Returns:
        A ValueStd named tuple of (value, std), with the associated expected
        value and standard deviation.
    """
    ys = np.asarray(ys)
    eta = ys[1:] / ys[:-1]
    N = len(eta)
    mean_eta = eta.mean()
    delta = np.log(mean_eta) - np.log(eta).mean()

    mean_r = (mean_eta + 2 * delta / N) / (1 + 2 * delta / N)
    denom = N / 2 - 2 * delta
    if denom > 0:
        std_r = np.sqrt(delta / denom) * mean_r
    else:
        std_r = float("inf")
    return ValueStd(mean_r, std_r)


def growth_factors(data):
    """
    Compute growth factors for each column of ys in data frame.

    Return a data frame with ["value", "std"] columns with the original columns
    in the index.
    """
    growth_factors = {}
    for key, col in data.items():
        col = trim_zeros(col)
        growth_factors[key] = growth_factor(col)

    return pd.DataFrame(growth_factors, index=["value", "std"]).T


def exponential_extrapolation(ys: Sequence, n: int, append=False) -> np.ndarray:
    """
    Receive a sequence  and return the next n points of the series
    extrapolating from the input data.

    Args:
        ys:
            Input data.
        n:
            Number of points of the extrapolation.
        append:
            If True, returns a concatenation of the input series with the
            extrapolation.
    """
    ys = np.asarray(ys)
    r, dr = growth_factor(ys)
    K = np.log(r)
    extrapolation = ys[-1] * np.exp(K * np.arange(1, n + 1))
    if append:
        return np.concatenate([ys, extrapolation])
    return extrapolation


def R0_from_cases(model, cases, params, method="OLS") -> ValueStd:
    """
    Read curve of cases and adjust the model R0 from cases.
    """
    # Methods that infer the growth ratio between successive observations
    if method.startswith("ratio-"):
        r, dr = growth_ratio_from_cases(cases, method=method[6:])

        R0 = formulas.R0_from_K(model, params, K=np.log(r))
        R0_plus = formulas.R0_from_K(model, params, K=np.log(r - min(dr, 0.9 * r)))
        R0_minus = formulas.R0_from_K(model, params, K=np.log(r + dr))

    # Methods that infer the exponential growth factor
    elif method in ("OLS",):
        K, dK = growth_factor_from_cases(cases, method=method)

        R0 = formulas.R0_from_K(model, params, K=K)
        R0_plus = formulas.R0_from_K(model, params, K=K - dK)
        R0_minus = formulas.R0_from_K(model, params, K=K + dK)

    else:
        raise ValueError(f"invalid method: {method!r}")

    dR0 = abs(R0_plus - R0_minus) / 2
    return ValueStd(R0, dR0)


def growth_ratio_from_cases(curves, method="GGBayes", **kwargs) -> ValueStd:
    """
    Return the growth rate combining the "cases" and "deaths" columns of an
    epidemic curve.

    Args:
        curves:
            A DataFrame with "cases" and "deaths" columns.
        method:
            Statistical method used to infer growth ratio.
            * GGBayes - A Bayesian Gamma distributed Growth rate.

    Keyword Args:
        Additional keyword arguments are passed to the smoothed_diff function.

    See Also:
        :func:`pydemic.fitting.smoothed_diff`
    """

    if method == "GGBayes":
        fn = lambda col: clean_exponential(curves[col], diff=True, **kwargs)
        cases, deaths = map(fn, curves.columns)
        ratios = [growth_factor(cases), growth_factor(deaths)]
        return average_inference(ratios)
    else:
        raise ValueError


def growth_factor_from_cases(curves, method="OLS", **kwargs) -> ValueStd:
    """
    Return the growth rate combining the "cases" and "deaths" columns of an
    epidemic curve.

    Args:
        curves:
            A DataFrame with "cases" and "deaths" columns.
        method:
            Statistical method used to infer the growth factor.
            * 'OLS' - ordinary least squares.

    Keyword Args:
        Additional keyword arguments are passed to the smoothed_diff function.

    See Also:
        :func:`pydemic.fitting.smoothed_diff`
    """

    if method == "OLS":

        def stats(col):
            data = curves[col]
            data = clean_exponential(data, diff=True, **kwargs)
            Y = np.log(data / data[0])
            X = np.arange(len(Y))
            ols = sm.OLS(Y, sm.add_constant(X))
            res = ols.fit()
            _, K = res.params
            ci = res.conf_int()
            dK = (ci[1, 1] - ci[1, 0]) / 2
            return ValueStd(K, dK)

        cases, deaths = map(stats, curves.columns)
        return average_inference([cases, deaths])
    else:
        raise ValueError



import operator
from typing import TYPE_CHECKING

from .utils import map_models, map_method, model_group_method, group

if TYPE_CHECKING:
    from .model_group import ModelGroup


class ModelGroupProp:
    """
    Base class for all model group properties.
    """

    group: "ModelGroup"
    prop_name: str = None
    MODEL_METHODS = frozenset()
    PROP_METHODS = frozenset()

    def __init__(self, group):
        self.group = group

    def __getitem__(self, item):
        models = self.iter_items(item)
        return map_models(operator.itemgetter(item), models)

    def __getattr__(self, item):
        from .model_group import ModelGroup

        if item in self.MODEL_METHODS:

            def method(*args, **kwargs):
                return ModelGroup(fn(*args, **kwargs) for fn in self.iter_attr(item))

            method.__name__ = item
            return method

        if item in self.PROP_METHODS:

            def method(*args, **kwargs):
                return [fn(*args, **kwargs) for fn in self.iter_attr(item)]

            method.__name__ = item
            return method

        return list(self.iter_attr(item))

    def __dir__(self):
        try:
            prop = next(self.iter_pros())
            prop_attrs = dir(prop)
        except IndexError:
            prop_attrs = ()
        return list({*super().__dir__(), *prop_attrs})

    def iter_props(self):
        """
        Iterate over each accessor.
        """
        prop_name = self.prop_name
        for m in self.group:
            yield getattr(m, prop_name)

    def iter_items(self, item):
        """
        Iterate over each accessor, selecting the given item.
        """
        for prop in self.iter_props():
            yield prop[item]

    def iter_attr(self, attr):
        """
        Iterate over each accessor, selecting the given attribute.
        """
        for prop in self.iter_props():
            yield getattr(prop, attr)

    def iter_method(self, *args, **kwargs):
        """
        Iterate over each accessor, executing the given method forwarding
        the passed arguments.
        """
        method, *args = args
        for fn in self.iter_attr(method):
            yield fn(*args, **kwargs)


class ModelGroupInfo(ModelGroupProp):
    """
    Implements the "info" attribute of model groups.
    """

    prop_name = "info"


class ModelGroupResults(ModelGroupProp):
    """
    Implements the "results" attribute of model groups.
    """

    prop_name = "results"


class ModelGroupClinical(ModelGroupProp):
    """
    Implements the "clinical" attribute of model groups.
    """

    prop_name = "clinical"
    MODEL_METHODS = {"clinical_model", "crude_model", "delay_model", "overflow_model"}

    def __call__(self, *args, **kwargs):
        return group(map_method("clinical", self.group, *args, **kwargs))

    crude_model = model_group_method("clinical.crude_model", out=group)
    delay_model = model_group_method("clinical.delay_model", out=group)
    overflow_model = model_group_method("clinical.overflow_model", out=group)



"""
Pydemic

Epidemiological calculator tuned specifically for COVID-19.
"""
# flake8: noqa
__version__ = "0.1.10"
__author__ = "Fábio Mendes"

# noinspection PyUnresolvedReferences
from . import models
from . import region
from . import formulas
from . import config
from . import db
from . import utils
from . import diseases
from . import clinical_models
from . import fitting
from .diseases import disease
from .params import Params, param, get_param, select_param
from .model_group import ModelGroup
from .logging import log



from functools import singledispatch

import numpy as np
import pandas as pd


@singledispatch
def to_json(obj):
    """
    Convert object to JSON.
    """
    try:
        method = obj.to_json
    except AttributeError:
        name = type(obj).__name__
        raise TypeError(f"'{name}' objects cannot be converted to JSON")
    else:
        return method()


@to_json.register(int)
@to_json.register(float)
@to_json.register(str)
@to_json.register(bool)
@to_json.register(type(None))
def _(x):
    return x


@to_json.register(list)
@to_json.register(tuple)
@to_json.register(set)
@to_json.register(frozenset)
@to_json.register(np.ndarray)
def _(xs):
    return [to_json(x) for x in xs]


@to_json.register(dict)
@to_json.register(pd.Series)
def _(xs):
    return {check_string(k): to_json(v) for k, v in xs.items()}


def check_string(x):
    if isinstance(x, str):
        return x
    elif isinstance(x, bool):
        return str(x).lower()
    elif isinstance(x, (int, float)):
        return str(x)
    elif x is None:
        return "null"
    elif x is ...:
        return "..."
    else:
        name = type(x).__name__
        raise TypeError(f"invalid type for object key: {name}")



import numpy as np
import pandas as pd

from .. import utils

elementwise = lambda fn: lambda x, i: x.iloc[i or ALL].apply(fn)
astype = lambda fn: lambda x, i: x.iloc[i or ALL].astype(fn)

ALL = slice(None, None)
DATA_TRANSFORMS = {}
MODEL_TRANSFORMS = {}
TIME_CONVERSIONS = {"days": 1, "weeks": 7, "months": 365.25 / 12, "years": 365.25}
RATIOS_PER_POP = {"pp": 1, "ppc": 100, "p1k": 1e3, "p10k": 1e4, "p100k": 1e5, "p1m": 1e6}


#
# Transforms and utility functions
#
def method(name, *args, **kwargs):
    """
    Call column method.
    """

    def fn(x, i):
        if i is not None:
            x = x.iloc[i]

        method = getattr(x, name)
        return method(*args, **kwargs)

    return fn


def peak_date(model, col, idx):
    """
    Date with the largest value of col.
    """
    return model.to_date(peak_time(model, col, idx))


def peak_time(model, col, idx):
    """
    Time with the largest value of col.
    """
    idx = np.argmax(model[col, idx])
    return model.times[idx]


def to_dataframe(model, col, idx):
    """
    Force columns to be data frames, even when results are vectors
    """
    name_, _, _ = col.partition(":")
    series = model[col, idx]
    if isinstance(series, pd.DataFrame):
        return series
    return pd.DataFrame({name_: series.values}, index=series.index)


def to_dates(model, col, idx):
    """
    Convert column index to use dates.
    """
    data = model[col, idx]
    data.index = model.to_dates(data.index)
    return data


def ratio_transform(factor):
    """
    Factory function for ratios per population transformers.
    """
    return lambda model, col, idx: factor * model[col, idx] / model.population


def time_transform(factor):
    """
    Divide time index by the given factor.
    """

    def fn(model, col, idx):
        data = model[col, idx]
        data.index = data.index / factor
        return data

    return fn


def initial(x, idx):
    """
    Get first value of series.
    """
    if idx is None:
        return x.iloc[0]
    return x.iloc[idx.start or 0]


def final(x, idx):
    """
    Get final value of series.
    """
    if idx is None:
        return x.iloc[-1]
    return x.iloc[utils.coalesce(idx.end, -1)]


#
# Update transform dictionaries
#
DATA_TRANSFORMS.update(
    {
        #
        # Simple queries
        "initial": initial,
        "final": final,
        "max": method("max"),
        "min": method("min"),
        #
        # Type conversions
        "np": lambda x, i: np.asarray(x) if i is None else np.asarray(x.iloc[i]),
        "int": astype(int),
        "float": astype(float),
        "str": astype(str),
        #
        # Elementwise transforms
        "round": elementwise(lambda x: round(x)),
        "round1": elementwise(lambda x: round(x, 1)),
        "round2": elementwise(lambda x: round(x, 2)),
        "round3": elementwise(lambda x: round(x, 3)),
        "human": elementwise(utils.fmt),
        "pcfmt": elementwise(utils.pc),
        "p1kfmt": elementwise(utils.pm),
        "p10kfmt": elementwise(utils.p10k),
        "p100kfmt": elementwise(utils.p100k),
    }
)
DATA_TRANSFORMS.update(first=DATA_TRANSFORMS["initial"], last=DATA_TRANSFORMS["final"])
MODEL_TRANSFORMS.update(
    {
        "peak-date": peak_date,
        "peak-time": peak_time,
        "df": to_dataframe,
        "dates": to_dates,
        #
        # Ratios per population
        **{k: ratio_transform(v) for k, v in RATIOS_PER_POP.items()},
        #
        # Time intervals
        **{k: time_transform(v) for k, v in TIME_CONVERSIONS.items()},
    }
)



import re
from typing import Type

import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_series_equal

from pydemic.clinical_models import CrudeFR, HospitalizationWithDelay
from pydemic import models

MODEL_CLASSES = [
    models.eSIR,
    # models.eSEIR,
    # models.eSEAIR,
    models.SIR,
    models.SEIR,
    models.SEAIR,
]


@pytest.fixture(params=MODEL_CLASSES, scope="session")
def model_cls(request) -> Type[models.Model]:
    return request.param


@pytest.fixture
def model(model_cls) -> models.Model:
    return model_cls()


class ModelTester:
    @pytest.fixture
    def m(self, model):
        return model

    @pytest.fixture(scope="session")
    def cls(self, model_cls):
        return model_cls


class TestInfectiousModels(ModelTester):
    def test_model_basic_api_interactions(self, m):
        m.run(7)
        assert m.R0 > 1
        assert m.name

        # Basic invariants
        assert m.iter == len(m.data)
        assert all(m.times == m.data.index)

        # Test the infectious column
        I = m["I"]
        assert_series_equal(I, m["infectious"])
        assert any(I > 0.0)
        assert I.dtype == float
        assert isinstance(I, pd.Series)
        assert all(m["I:dates"].index == m.dates)

        # Test item accessors
        assert m["I", 0] == m.data.loc[0, "infectious"]
        assert m["I", -1] == m.data.loc[len(m.data) - 1, "infectious"]

    def test_can_override_model_parameters_during_initialization(self, cls):
        m = cls(R0=2)
        assert m.R0 == 2.0

        m.run(3)
        assert m.R0 == 2.0

    def test_model_can_run_during_initialization(self, cls):
        m1 = cls()
        m1.run(10)
        m2 = cls(run=10)
        assert all(m1.data == m2.data)

    def test_model_time_is_not_reset_from_clinical_model(self, m):
        m.run(10)
        date = m.date
        h = m.clinical()
        assert h.time == m.time == 10
        assert h.date == m.date == date

    def test_model_can_run_multiple_times(self, m):
        m.run(5)
        m.run(5)
        m.run(5)
        assert m.time == 15
        assert len(m.times) == 16
        assert len(set(m.times)) == 16
        assert all(np.diff(m.times) > 0)

    def test_model_data_uses_numeric_index(self, m):
        assert m.data.index.dtype == float

    def test_model_exposes_parameters_as_timeseries(self, m):
        m.run(7)
        R0 = m.R0
        series = m["R0"]
        print(series)
        assert len(series) == 8
        assert (series == R0).all()

    def test_clinical_accessor(self, m):
        h = m.clinical()
        assert h.empirical_CFR < m.exposed
        assert h.empirical_IFR < m.exposed
        assert isinstance(h, CrudeFR)
        assert type(h) == type(m.clinical.crude_model())
        assert isinstance(m.clinical.delay_model(), HospitalizationWithDelay)

    def test_clinical_model_basic_api(self, m):
        m.run(10)
        cm = m.clinical()
        c = cm["cases"]
        d = cm["deaths"]
        h = cm["hospitalized"]
        H = cm["hospitalized_cases"]
        assert all(h <= H)
        assert all(d <= H)
        assert all(H <= c)


class TestGetitemInteface:
    def test_transforms(self):
        m = models.SIR(R0=3, infectious_period=0.5, run=10, date="1970-01-01")

        I = m["I"]
        first, last = I.iloc[[0, -1]]
        assert_series_equal(m["infectious"], I)
        assert_series_equal(m["infectious:int"], m["I:int"])

        # Type conversions
        assert m["I:int"].dtype == int
        assert m["I:round"].dtype == int
        assert m["I:int:max"] >= 250_000
        assert_pattern(r"\d+", m["I:round"])
        assert_pattern(r"\d+\.\d", m["I:round1"])
        assert_pattern(r"\d+\.\d{1,2}", m["I:round2"])
        assert_pattern(r"\d+\.\d{1,3}", m["I:round3"])

        # Simple transforms
        assert m["I:initial"] == first
        assert m["I:final"] == last
        assert m["I:max"] >= 250_000
        assert m["I:min"] == 1.0
        assert m["I:peak-time"] == 4.0
        assert m["I:peak-date"] == pd.to_datetime("1970-01-05")
        assert type(m["I:np"]) is np.ndarray
        assert m["I:np:min"] == 1.0
        assert m["I:str:initial"] == "1.0"

        # Rate conversions
        assert (m["I:pp"] <= 1).all()
        assert (m["I:ppc"] <= 100).all()
        assert (m["I:p1k"] <= 1e3).all()
        assert (m["I:p10k"] <= 1e4).all()
        assert (m["I:p100k"] <= 1e5).all()
        assert (m["I:p1m"] <= 1e6).all()
        assert m["I:ppc:max"] >= 25


def assert_pattern(regex, data):
    pattern = re.compile(regex)
    check = pattern.fullmatch
    for i, x in enumerate(data):
        if check(str(x)) is None:
            msg = f"invalid pattern found in pos {i}: {x!r}"
            raise AssertionError(msg)



import pandas as pd
import pytest
from pandas.testing import assert_frame_equal
from pytest import approx

from pydemic.api.covid19 import epidemic_curve
from pydemic.diseases import covid19


class TestCovid19:
    def test_covid19_reference_tables(self):
        for source in [None, "verity"]:
            assert isinstance(covid19.mortality_table(source), pd.DataFrame)
        for source in [None, "verity"]:
            assert isinstance(covid19.hospitalization_table(source), pd.DataFrame)

    def test_covid19_default_reference_tables(self):
        table = covid19.mortality_table
        assert_frame_equal(table(), table("verity"))

        table = covid19.hospitalization_table
        assert_frame_equal(table(), table("verity"))

    def test_covid_dict_params(self):
        data = covid19.to_dict()
        res = {
            "R0": 2.74,
            "rho": 0.45,
            #
            "case_fatality_ratio": 0.011,
            "infection_fatality_ratio": 0.00566,
            #
            "icu_fatality_ratio": 0.490,
            "hospital_fatality_ratio": 0.141,
            #
            "incubation_period": 3.69,
            "infectious_period": 3.47,
            #
            "hospitalization_period": 10.0,
            "icu_period": 7.5,
            #
            "critical_period": 7.5,
            "severe_period": 10.0,
            #
            "prob_critical": 0.0225,
            "prob_severe": 0.0778,
            "prob_symptoms": 0.532,
            #
            "hospitalization_overflow_bias": 0.25,
        }
        for k, v in res.items():
            assert data[k] == approx(v, rel=0.01, abs=1e-3), k
        assert set(res.keys()) == set(data.keys())
        assert data == covid19.to_json()

    def _test_params(self):
        p = covid19.params()

        # Aliases
        assert p.Qs == p.prob_symptoms
        assert p.Qsv == p.prob_severe
        assert p.Qcr == p.prob_critical

        # Transforms
        assert p.gamma == 1 / p.infectious_period
        assert p.sigma == 1 / p.incubation_period


class TestCovid19APIs:
    @pytest.mark.slow
    @pytest.mark.external
    def test_corona_api(self):
        br = epidemic_curve("BR", api="corona-api.com", extra=True)
        assert list(br.columns) == ["cases", "deaths", "recovered"]

        br = epidemic_curve("BR", api="corona-api.com", extra=False)
        assert list(br.columns) == ["cases", "deaths"]

    @pytest.mark.slow
    @pytest.mark.external
    def test_brasil_io_api(self):
        df = epidemic_curve("BR-DF", api="brasil.io")
        assert list(df.columns) == ["cases", "deaths"]



from typing import (
    TYPE_CHECKING,
    Type,
    Tuple,
    Iterator,
    Dict,
    FrozenSet,
    Iterable,
    TypeVar,
    Sequence,
)

from .params_info import ParamsInfo
from .. import utils

if TYPE_CHECKING:
    from ..models import Model  # noqa: F401
    from ..models.metaclass import ModelMeta

T = TypeVar("T")


class Meta:
    """
    Meta information about model.

    Attributes:
        params:
            Information about parameters, subclass of :cls:`ParamsInfo`
    """

    cls: Type["Model"]
    model_name: str
    params: "ParamsInfo"
    variables: Tuple[str]
    data_aliases: Dict[str, str]
    plot_columns: FrozenSet[str]
    ndim: int

    @classmethod
    def from_arguments(cls, kind, meta):
        """
        Create Meta instance for the given kind from the list of base classes and
        an inner Django-like Meta class declaration.
        """

        if meta is not None:
            ns = vars(meta)
            kwargs = {k: v for k, v in ns.items() if not k.startswith("_")}
        else:
            kwargs = {}
        return Meta(kind, **kwargs)

    def __init__(self, cls, **kwargs):
        cls.meta = self
        self.cls = cls
        self.explicit_kwargs = kwargs.copy()
        self.variables = tuple(p.name for p in sorted(iter_state_variables(cls)))
        self.ndim = len(self.variables)
        self.params = ParamsInfo(cls)
        self.data_aliases = data_aliases(cls, kwargs.pop("data_aliases", None))

        # Keyword variables
        keywords = explicit_keywords(cls, self=True)
        self.model_name = keywords.get("model_name", "Model")

        # Plot columns
        new = kwargs.pop("plot_columns", (...,))
        bases = get_base_meta_attr(cls, "plot_columns", ())
        self.plot_columns = frozenset(merge_with_ellipsis(new, bases))

        # Check keyword args
        invalid = set(kwargs) - {"model_name", "plot_columns"}
        if invalid:
            raise TypeError(f"invalid arguments: {invalid}")

    def get_variable_index(self, name):
        """
        Return index from variable name or alias.
        """
        name = self.data_aliases.get(name, name)
        return self.variables.index(name)

    def __repr__(self):
        return f"<{self.cls.__name__}.meta object>"


def explicit_keywords(cls, self=False):
    """
    Extract explicit keyword arguments from meta object from all parent types.
    """

    kwargs = {}
    metas = list(iter_metas(cls, self=self))

    for meta in reversed(metas):
        kwargs.update(meta.explicit_kwargs)

    return kwargs


def iter_state_variables(cls) -> Iterator[utils.state_property]:
    """
    Return a list of state variables for class.
    """
    for attr in dir(cls):
        value = getattr(cls, attr, None)
        if isinstance(value, utils.state_property):
            yield value


def iter_metas(cls: "ModelMeta", self=False):
    """
    Iterate over _meta objects of parent classes.
    """
    if self:
        yield cls.meta

    for base in cls.mro()[1:]:
        if hasattr(base, "meta"):
            yield base.meta


def get_base_meta_attr(cls, prop, default):
    """
    Return attribute from the _meta attribute of a base class.
    """

    for meta in iter_metas(cls):
        try:
            return meta.explicit_kwargs[prop]
        except KeyError:
            pass
    return default


def data_aliases(cls, extra=None):
    """
    Return the data_aliases meta attribute for the given class.

    Extra is an optional dictionary with additional data alias definitions.
    """
    out = {}
    for base in reversed(cls.__bases__):
        try:
            out.update(base.meta.data_aliases)
        except AttributeError:
            continue

    out.update(extra or ())

    # Remove null aliases
    for k, v in list(out.items()):
        if v is None:
            del out[k]
    return out


def merge_with_ellipsis(xs: Iterable[T], extra: Sequence[T]) -> Iterator[T]:
    """
    Yield xs, but extra whenever encounters an ellipsis in xs.
    """
    for x in xs:
        if x is ...:
            yield from extra
        else:
            yield x



import re
import warnings
from abc import ABC
from pathlib import Path

import numpy as np
import pandas as pd
import sidekick as sk
from sidekick import X

import mundi
from pydemic.utils import trim_zeros, force_monotonic
from .disease_params import DiseaseParams
from .utils import (
    QualDataT,
    read_table,
    Dataset,
    QualValueT,
    world_age_distribution,
    age_adjusted_average,
    set_age_distribution_default,
    estimate_real_cases,
)
from .. import db
from .. import fitting as fit
from ..config import set_cache_options
from ..types import ImproperlyConfigured
from ..utils import to_json

# Types
set_cache_options("diseases-api", compress=True)


class Disease(ABC):
    """
    Basic interface that exposes information about specific diseases.
    """

    # Attributes
    name: str = ""
    description: str = ""
    full_name: str = ""
    path: str = sk.lazy(lambda _: "diseases/" + _.name.lower().replace(" ", "-"))
    abspath: Path = sk.lazy(lambda _: db.DATABASES / _.path)
    _default_params = sk.lazy(lambda _: DiseaseParams(_))

    # Constants
    MORTALITY_TABLE_DEFAULT = "default"
    MORTALITY_TABLE_ALIASES = {}
    MORTALITY_TABLE_DESCRIPTIONS = {}
    HOSPITALIZATION_TABLE_DEFAULT = "default"
    HOSPITALIZATION_TABLE_ALIASES = {}
    HOSPITALIZATION_TABLE_DESCRIPTIONS = {}
    PARAMS_BLACKLIST = frozenset({"to_json", "to_record", "to_dict", "params", "epidemic_curve"})

    def __init__(self, name=None, description=None, full_name=None, path=None):
        self.name = name or self.name or type(self).__name__
        self.full_name = full_name or self.full_name
        self.description = description or self.description or self.__doc__.strip()
        self.path = path or self.path

    def __str__(self):
        return self.full_name

    def __repr__(self):
        return f"{type(self).__name__}()"

    def __bool__(self):
        return True

    def mortality_table(
        self, source: str = None, qualified=False, extra=False, region=None
    ) -> QualDataT:
        """
        Return the mortality table of the disease.

        The mortality table is stratified by age and has at least two
        columns, one with the CFR and other with the IFR per age group.

        Args:
            source (str):
                Select source/reference that collected this data. Different
                datasets may be available representing the best knowledge by
                different teams or assumptions about the disease. This argument
                is a string identifier for that version of the data.
            qualified (bool):
                If True, return a :cls:`Dataset` namedtuple with
                (data, source, notes) attributes.
            extra (bool):
                If True, display additional columns alongside with ["IRF", "CFR"].
                The extra columns can be different for each dataset.
        """
        df = self._read_dataset("mortality-table", source, qualified)
        return df if extra else df[["IFR", "CFR"]]

    def hospitalization_table(
        self, source=None, qualified=False, extra=False, region=None
    ) -> QualDataT:
        """
        Return the hospitalization table of the disease. The hospitalization
        table is stratified by age and has two columns: "severe" and "critical".
        Each column represents the probability of developing either "severe"
        (requires clinical treatment) or "critical" (requires ICU) cases.

        In both cases, the probabilities are interpreted as the ratio between
        hospitalization and cases, *excluding* asymptomatic individuals

        Args:
            source (str):
                Select source/reference that collected this data. Different
                datasets may be available representing the best knowledge by
                different teams or assumptions about the disease. This argument
                is a string identifier for that version of the data.
            qualified (bool):
                If True, return a :cls:`Dataset` namedtuple with
                (data, source, notes) attributes.
            extra (bool):
                If True, display additional columns alongside with ["severe",
                "critical"]. The extra columns can be different for each dataset.
        """
        df = self._read_dataset("hospitalization-table", source, qualified)
        return df if extra else df[["severe", "critical"]]

    def _read_dataset(self, which, source, qualified) -> QualDataT:
        """
        Worker method for many data reader methods.
        """
        attr = which.replace("-", "_").upper()
        aliases = getattr(self, attr + "_ALIASES")
        descriptions = getattr(self, attr + "_DESCRIPTIONS")
        source = source or getattr(self, attr + "_DEFAULT")

        if ":" in source:
            name, _, arg = source.partition(":")
            name = aliases.get(name, name)
            method = re.sub(r"[\s-]", "_", f"get_data_{which}_{name}")
            try:
                fn = getattr(self, method)
            except AttributeError:
                raise ValueError(f"invalid method source: {source!r}")
            else:
                data = fn(arg)
                description = fn.__doc__
        else:
            name = aliases.get(source, source)
            description = descriptions.get(name, "").strip()
            data = read_table(f"{self.path}/{which}-{name}").copy()

        if qualified:
            return Dataset(data, source, description)
        return data

    def case_fatality_ratio(self, **kwargs) -> QualValueT:
        """
        Compute the case fatality ratio with possible age distribution
        adjustments.

        Args:
            age_distribution:
                A age distribution or a string with a name of a valid Mundi
                region with known demography. If not given, uses the average
                world age distribution.
            source:
                Reference source used to provide the mortality table.
        """
        return self._fatality_ratio("CFR", **kwargs)

    def infection_fatality_ratio(self, **kwargs) -> QualValueT:
        """
        Compute the infection fatality ratio with possible age distribution
        adjustments.

        Args:
            age_distribution:
                A age distribution or a string with a name of a valid Mundi
                region with known demography. If not given, uses the average
                world age distribution.
            source:
                Reference source used to provide the mortality table.
        """
        return self._fatality_ratio("IFR", **kwargs)

    def _fatality_ratio(self, col, age_distribution=None, source=None, region=None):
        table = self.mortality_table(source=source)

        if age_distribution is None and region:
            ages = mundi.region(region).age_distribution
        elif age_distribution is None:
            ages = world_age_distribution()
        else:
            ages = age_distribution
        return age_adjusted_average(ages, table[col])

    def epidemic_curve(
        self,
        region,
        diff=False,
        smooth=False,
        real=False,
        keep_observed=False,
        window=14,
        trim_empty="left",
        keep_reversions=False,
        **kwargs,
    ) -> pd.DataFrame:
        """
        Load epidemic curve for the given region.

        Args:
            region:
                Mundi r/egion or a string with region code.
            diff (bool):
                If diff=True, return the number of new daily cases instead of
                the accumulated epidemic curves.
            smooth (bool):
                If True, return a data frame with raw and smooth columns. The
                resulting dataframe adopts a MultiIndex created from the
                product of ["observed", "smooth"] x ["cases", "deaths"]
            real (bool, str):
                If True, estimate the real number of cases by trying to correct
                for some ascertainment rate.
            keep_observed (bool):
                If True, keep the raw observed cases under the "cases_observed"
                and "deaths_observed" columns.
            window (int):
                Size of the triangular smoothing window.
            keep_reversions (bool):
                If True, prevent the default behavior of cleaning data that
                violates a monotonic increasing behavior.
            trim_empty (str):
                Direction to trim rows with only null values. By default, it
                just trims the left hand side of the series (i.e., older entries).
                This value can be either 'left', 'right', 'both', or 'none'.

        """
        data = self._epidemic_curve(mundi.region(region), **kwargs)
        data = trim_zeros(data, trim_empty)
        if not keep_reversions:
            data = force_monotonic(data)
        if real:
            method = "CFR" if real is True else real
            params = self.params(region=region)
            real = estimate_real_cases(data, params, method)
            if keep_observed:
                rename = {"cases": "cases_observed", "deaths": "deaths_observed"}
                data = pd.concat([real, data.rename(rename, axis=1)], axis=1)
            else:
                data = real

        if diff:
            values = np.diff(data, prepend=0, axis=0)
            data = pd.DataFrame(values, index=data.index, columns=data.columns)

        if smooth:
            columns = pd.MultiIndex.from_product([["observed", "smooth"], data.columns])
            data = pd.concat([data, fit.smooth(data, window)], axis=1)
            data.columns = columns

        return data

    def _epidemic_curve(self, region, **kwargs):
        raise NotImplementedError

    #
    # Basic epidemiology
    #
    def R0(self, **kwargs) -> QualValueT:
        """
        R0 for disease.
        """
        return NotImplemented

    def rho(self, **kwargs):
        """
        Relative transmissibility of asymptomatic cases.

        The default value is 1.0. This value makes the distinction between
        symptomatic and asymptomatic as a mere notification problem
        """
        return 1.0

    #
    # Clinical progression
    #

    # Probabilities
    def prob_severe(self, **kwargs) -> QualValueT:
        """
        Probability that a case become severe enough to recommend hospitalization.

        Estimated from "severe cases / symptomatic cases".
        """
        return self._prob_from_hospitalization_table("severe", **kwargs)

    def prob_critical(self, **kwargs) -> QualValueT:
        """
        Probability that a case become severe enough to recommend intensive
        care treatment.

        Estimated from "critical cases / symptomatic cases".
        """
        try:
            return self._prob_from_hospitalization_table("critical", **kwargs)
        except KeyError:
            pass
        try:
            return self.CFR(**kwargs) / self.icu_fatality_ratio(**kwargs)
        except RecursionError:
            model = type(self).__name__
            raise ImproperlyConfigured(
                f"""
{model} class must implement either prob_critical() or icu_fatality_ratio()
methods. Otherwise, the default implementation creates a recursion between both
methods."""
            )

    def _prob_from_hospitalization_table(self, col, **kwargs):
        """
        Worker function for prob_critical and prob_severe methods.
        """
        ages = set_age_distribution_default(kwargs, drop=True)
        values = self.hospitalization_table(**kwargs)
        return age_adjusted_average(ages, values[col])

    def prob_aggravate_to_icu(self, **kwargs) -> QualValueT:
        """
        Probability that an hospitalized patient require ICU.
        """
        return self.prob_critical(**kwargs) / self.prob_severe(**kwargs)

    def prob_symptoms(self, **kwargs) -> QualDataT:
        """
        Probability that disease develop symptomatic cases.
        """
        e = 1e-50
        ages = set_age_distribution_default(kwargs, drop=True)
        table = self.mortality_table(**kwargs)
        ratios = (table["IFR"] + e) / (table["CFR"] + e)
        return age_adjusted_average(ages, ratios)

    def hospitalization_overflow_bias(self, **kwargs) -> QualValueT:
        """
        Increase in the fraction of critical patients when severe patients are
        not treated in a proper healthcare facility.

        The default implementation ignores this phenomenon and returns zero.
        """
        return 0.0

    # Durations
    def infectious_period(self, **kwargs) -> QualValueT:
        """
        Period in which cases are infectious.
        """
        return NotImplemented

    def incubation_period(self, **kwargs) -> QualValueT:
        """
        Period between infection and symptoms onset.
        """
        return NotImplemented

    def severe_period(self, **kwargs) -> QualValueT:
        """
        Duration of the "severe" state of disease.

        The default implementation assumes it is the same as the hospitalization
        period. If this method has a different implementation, this parameter
        is interpreted as the duration of the severe state in the absence of
        hospitalization.
        """
        return self.hospitalization_period(**kwargs)

    def critical_period(self, **kwargs) -> QualValueT:
        """
        Duration of the "critical" state of disease.

        The default implementation assumes it is the same as the ICU
        period. If this method has a different implementation, this parameter
        is interpreted as the duration of the severe state in the absence of
        hospitalization.
        """
        return self.icu_period(**kwargs)

    def icu_period(self, **kwargs) -> QualValueT:
        """
        Duration of ICU treatment.

        The default implementation assumes this to be zero, meaning the disease
        never progress to a critical state or kills instantly if this state
        is reached.

        Obviously, most diseases would need to override this method.
        """
        return 0.0

    def hospitalization_period(self, **kwargs) -> QualValueT:
        """
        Duration of hospitalization treatment.

        The default implementation assumes this to be zero, meaning the disease
        never progress to a severe state or kills instantly if this state
        is reached.

        Obviously, most diseases would need to override this method.
        """
        return 0.0

    # Delays
    def symptom_delay(self, **kwargs):
        """
        Average duration between becoming infectious and symptom delay.
        """
        return 0.0

    def critical_delay(self, **kwargs) -> QualValueT:
        """
        Average duration between symptom onset and necessity of ICU admission.
        """
        return NotImplemented

    def severe_delay(self, **kwargs) -> QualValueT:
        """
        Average duration between symptom onset and necessity of hospital admission.
        """
        return NotImplemented

    def death_delay(self, **kwargs):
        """
        Average duration between symptom onset and necessity of hospital
        admission.
        """
        return self.critical_delay(**kwargs) + self.icu_period(**kwargs)

    # Derived values
    def gamma(self, **kwargs):
        """
        The inverse of infectious period.
        """
        return 1 / self.infectious_period(**kwargs)

    def sigma(self, **kwargs):
        """
        The inverse of incubation period.
        """
        return 1 / self.incubation_period(**kwargs)

    def hospital_fatality_ratio(self, **kwargs) -> QualValueT:
        """
        Probability of death once requires hospitalization.
        """
        # FIXME: make properly age-stratified
        return self.CFR(**kwargs) / self.prob_severe(**kwargs)

    def icu_fatality_ratio(self, **kwargs) -> QualValueT:
        """
        Probability of death once requires intensive care.

        The default implementation assumes that death occurs only after reaching
        a critical state.
        """
        # FIXME: make properly age-stratified
        return self.CFR(**kwargs) / self.prob_critical(**kwargs)

    # Aliases
    def Qs(self, **kwargs) -> QualValueT:
        """
        Alias to "prob_symptoms".
        """
        return self.prob_symptoms(**kwargs)

    def Qsv(self, **kwargs) -> QualValueT:
        """
        Alias to "prob_severe".
        """
        return self.prob_severe(**kwargs)

    def Qcr(self, **kwargs) -> QualValueT:
        """
        Alias to "prob_critical".
        """
        return self.prob_critical(**kwargs)

    def CFR(self, **kwargs) -> QualValueT:
        """
        Alias to "case_fatality_ratio"
        """
        return self.case_fatality_ratio(**kwargs)

    def IFR(self, **kwargs) -> QualValueT:
        """
        Alias to "infection_fatality_ratio"
        """
        return self.infection_fatality_ratio(**kwargs)

    def HFR(self, **kwargs) -> QualValueT:
        """
        Alias to "hospital_fatality_ratio"
        """
        return self.hospital_fatality_ratio(**kwargs)

    def ICUFR(self, **kwargs) -> QualValueT:
        """
        Alias to "icu_fatality_ratio"
        """
        return self.icu_fatality_ratio(**kwargs)

    #
    # Conversions to parameters
    #
    def params(self, *args, **kwargs) -> DiseaseParams:
        """
        Wraps disease in a parameter namespace.

        This is useful and a safer alternative to use disease as an argument
        to several functions that expect params.
        """
        if not args and not kwargs:
            return self._default_params
        return DiseaseParams(self, *args, **kwargs)

    def to_record(self, **kwargs) -> sk.record:
        """
        Return a sidekick record object with all disease parameters.
        """
        return sk.record(self.to_dict(**kwargs))

    def to_dict(self, *, alias=False, transform=False, **kwargs) -> dict:
        """
        Return a dict with all epidemiological parameters.
        """

        methods = (
            "R0",
            "rho",
            "case_fatality_ratio",
            "infection_fatality_ratio",
            "hospital_fatality_ratio",
            "icu_fatality_ratio",
            "infectious_period",
            "incubation_period",
            "hospitalization_period",
            "icu_period",
            "hospitalization_overflow_bias",
            "severe_period",
            "critical_period",
            "prob_symptoms",
            "prob_severe",
            "prob_critical",
        )

        aliases = (
            ("Qsv", "prob_severe"),
            ("Qcr", "prob_critical"),
            ("Qs", "prob_symptoms"),
            ("IFR", "infection_fatality_ratio"),
            ("CFR", "case_fatality_ratio"),
            ("HFR", "hospital_fatality_ratio"),
            ("ICUFR", "icu_fatality_ratio"),
        )

        transforms = (
            ("gamma", "infectious_period", (1 / X)),
            ("sigma", "incubation_period", (1 / X)),
        )

        out = {}
        for method in methods:
            try:
                out[method] = getattr(self, method)(**kwargs)
            except Exception as e:
                name = type(e).__name__
                warnings.warn(f"raised when calling disease.{method}():\n" f'    {name}: "{e}" ')
                raise

        if alias:
            for k, v in aliases:
                if v in out:
                    out[k] = out[v]

        if transform:
            for k, v, fn in transforms:
                if v in out:
                    out[k] = fn(out[v])

        return out

    def to_json(self, **kwargs) -> dict:
        """
        Similar to :meth:`to_dict`, but converts non-compatible values such as
        series and dataframes to json.
        """
        return to_json(self.to_dict(**kwargs))



from abc import ABC
from typing import Any

from sidekick import X
from sidekick import placeholder as _

from .model import Model
from ..packages import np, sk, integrate, pd
from ..utils import param_property, state_property, inverse_transform

Param = Any
NOT_GIVEN = object()


class AbstractSIR(Model, ABC):
    """
    Abstract base class for all SIR-based models.
    """

    class Meta:
        model_name = "SIR"
        data_aliases = {
            "S": "susceptible",
            "I": "infectious",
            "R": "recovered",
            "E": "infectious",
            "exposed": "infectious",
        }

    # Basic epidemiological parameters
    infectious_period: float = param_property("infectious_period", default=1.0)

    # Simulation state
    susceptible: float = state_property(0)
    exposed: float = sk.alias("infectious")  # an alias to infectious
    infectious: float = state_property(1)
    recovered: float = state_property(2)

    #
    # Accessors for parameters from other models
    #

    # SIR model assumes all cases are symptomatic
    prob_symptoms: float = 1.0
    Qs: float = sk.alias("prob_symptoms", mutable=True)

    # A null incubation period implies an infinite sigma. We instead assign
    # a very small value to avoid ZeroDivision errors
    incubation_period: float = 1e-50
    sigma: float = sk.alias("incubation_period", transform=(1 / X), prepare=(1 / X))

    # The rationale for Rho defaulting to 1 is that the difference between
    # regular and asymptomatic cases is simply a matter of notification
    rho: float = 1.0

    # Derived parameters and expressions
    gamma: float = inverse_transform("infectious_period")
    beta: float = sk.property(_.gamma * _.R0)
    K: float = sk.property(_.gamma * (_.R0 - 1))

    #
    # Model API
    #
    def _initial_state(self):
        return np.array((self.population - 1, 1, 0), dtype=float)

    def extrapolate_cases(self, which="infectious"):
        """
        Extrapolates how many cumulative cases where necessary to form the
        current number of infectious.
        """
        I = self[which]
        return I.iloc[0] * (self.beta + 1e-50) / max(self.K, 0.1)

    def _initial_infected(self):
        return self.extrapolate_cases()

    def _initial_cases(self, col="infectious"):
        return self.extrapolate_cases() * self.Qs

    #
    # Process data
    #
    def get_data_force(self, idx):
        I = self["infectious", idx]
        N = self["N", idx]
        return self.beta * (I / N)

    def get_data_resolved_cases(self, idx):
        I = self["infectious", idx]
        res = integrate.cumtrapz(I, self.times, initial=0.0)
        return pd.Series(res * self.gamma, index=I.index)

    def get_data_cases(self, idx):
        infections = self["force"] * self["susceptible"] * self.Qs
        res = integrate.cumtrapz(infections, self.times, initial=0.0)
        data = pd.Series(res + self.initial_cases, index=infections.index)
        return data[idx] if idx is not None else data

    def get_data_infected(self, idx):
        infections = self["force"] * self["susceptible"]
        res = integrate.cumtrapz(infections, self.times, initial=0.0)
        data = pd.Series(res + self.initial_infected, index=infections.index)
        return data[idx] if idx is not None else data



from abc import ABC

import sidekick as sk

from ..models import Model, ODEModel
from ..utils import param_property, param_alias


class ClinicalModel(Model, ABC):
    """
    Base class for clinical models that track the infection curve and
    models the clinical history of patients.
    """

    class Meta:
        model_name = "Clinical"
        data_aliases = {"H": "hospitalized", "D": "deaths"}
        plot_columns = ("hospitalized_cases", "hospitalized", "deaths")

    # Delegates (population parameters)
    population = sk.delegate_to("infection_model")
    K = sk.delegate_to("infection_model")
    disease = sk.delegate_to("infection_model")
    disease_params = sk.delegate_to("infection_model")
    region = sk.delegate_to("infection_model")
    age_distribution = sk.delegate_to("infection_model")
    age_pyramid = sk.delegate_to("infection_model")

    # Properties and aliases
    case_fatality_ratio = param_property(default=0.0)
    infection_fatality_ratio = param_property(default=lambda _: _.CFR)
    CFR = param_alias("case_fatality_ratio")
    IFR = param_alias("infection_fatality_ratio")

    @property
    def empirical_CFR(self):
        return self["empirical_CFR:final"]

    @property
    def empirical_IFR(self):
        return self["empirical_IFR:final"]

    @property
    def clinical_model(self):
        return self

    def __init__(self, infection_model, *args, **kwargs):
        self.infection_model = infection_model

        for k in ("disease", "region"):
            if k not in kwargs:
                kwargs[k] = getattr(infection_model, k)

        kwargs.setdefault("name", infection_model.name)
        super().__init__(*args, **kwargs)

    def __getattr__(self, item):
        try:
            return getattr(self.infection_model, item)
        except AttributeError:
            name = type(self).__name__
            raise AttributeError(f'"{name}" object has no "{item}" attribute')

    def copy(self, **kwargs):
        kwargs["infection_model"] = self.infection_model.copy()
        return super().copy(**kwargs)

    #
    # Data accessors
    #
    def get_column(self, name, idx):
        name = self.meta.data_aliases.get(name, name)
        try:
            return super().get_column(name, idx)
        except ValueError:
            return self.infection_model.get_column(name, idx)

    # Basic columns
    def get_data_population(self, idx):
        """
        Total population minus deaths.
        """
        return self.infection_model["population", idx] - self["deaths", idx]

    def get_data_infectious(self, idx):
        """
        Infectious population according to infectious model.

        This is usually the starting point of all clinical models.
        """
        return self.infection_model["infectious", idx]

    def get_data_cases(self, idx):
        """
        Cumulative curve of cases.

        A case is typically defined as an individual who got infected AND
        developed recognizable clinical symptoms.
        """
        return self.infection_model["cases", idx]

    def get_data_infected(self, idx):
        """
        Cumulative curve of infected individuals.

        Infected individuals might not develop clinical symptoms. They may never
        develop symptoms (asymptomatic) or develop them in a future time.
        """
        try:
            return self.infection_model["infected", idx]
        except KeyError:
            return self["cases", idx]

    # Derived methods
    def get_data_empirical_CFR(self, idx):
        """
        Empirical CFR computed as current deaths over cases.
        """
        return (self["deaths", idx] / self["cases", idx]).fillna(0.0)

    def get_data_empirical_IFR(self, idx):
        """
        Empirical IFR computed as current deaths over infected.
        """
        return (self["deaths", idx] / self["infected", idx]).fillna(0.0)

    # Abstract interface
    def get_data_death_rate(self, idx):
        """
        Daily number of deaths.
        """
        return self["deaths", idx].diff().fillna(0)

    def get_data_deaths(self, idx):
        """
        Cumulative curve of deaths.
        """
        raise NotImplementedError("must be implemented in sub-classes")

    def get_data_severe(self, idx):
        """
        Current number of severe cases.

        Severe cases usually require hospitalization, but have a low death risk.
        """
        raise NotImplementedError("must be implemented in sub-classes")

    def get_data_severe_cases(self, idx):
        """
        Cumulative number of severe cases.

        Severe cases usually require hospitalization, but have a low death risk.
        """
        raise NotImplementedError("must be implemented in sub-classes")

    def get_data_critical(self, idx):
        """
        Current number of critical cases.

        Critical cases require intensive care and are at a high risk of death.
        """
        raise NotImplementedError("must be implemented in sub-classes")

    def get_data_critical_cases(self, idx):
        """
        Cumulative number of critical cases.

        Critical cases require intensive care and are at a high risk of death.
        """
        raise NotImplementedError("must be implemented in sub-classes")

    def get_data_hospitalized(self, idx):
        """
        Cases currently occupying a hospital bed.

        In an ideal world, this would be equal to the number of severe cases.
        The default implementation assumes that.
        """
        return self["severe", idx]

    def get_data_hospitalized_cases(self, idx):
        """
        Cumulative number of hospitalizations.

        Default implementation assumes equal to the number of severe cases.
        """
        return self["severe_cases", idx]

    def get_data_icu(self, idx):
        """
        Number of ICU patients.

        In an ideal world, this would be equal to the number of critical cases.
        The default implementation assumes that.
        Default implementation assumes equal to the number of critical cases.
        """
        return self["critical", idx]

    def get_data_icu_cases(self, idx):
        """
        Cumulative number of ICU patients.

        Default implementation assumes equal to the number of critical cases.
        """
        return self["critical_cases", idx]

    #
    # Other functions
    #
    def plot(self, components=None, *, ax=None, logy=False, show=False, **kwargs):
        if components is None:
            self.infection_model.plot(**kwargs)
            components = self.meta.plot_columns
        super().plot(components, show=show, **kwargs)


class ClinicalObserverModel(ClinicalModel, ABC):
    """
    A clinical model that derives all its dynamic variables from simple
    transformations from the infected population.

    Observer models do not control directly their own dynamics and simply
    delegate time evolution to the corresponding infectious model.
    """

    DATA_COLUMNS = ()

    # Delegates (times and dates)
    times = sk.delegate_to("infection_model")
    dates = sk.delegate_to("infection_model")
    iter = sk.delegate_to("infection_model")
    time = sk.delegate_to("infection_model")
    date = sk.delegate_to("infection_model")
    state = sk.delegate_to("infection_model")

    # Methods delegates
    set_ic = sk.delegate_to("infection_model")
    set_data = sk.delegate_to("infection_model")
    set_cases = sk.delegate_to("infection_model")
    run = sk.delegate_to("infection_model")
    run_until = sk.delegate_to("infection_model")
    trim_dates = sk.delegate_to("infection_model")
    reset = sk.delegate_to("infection_model")
    epidemic_model_name = sk.delegate_to("infection_model")

    def __init__(self, model, params=None, *, date=None, **kwargs):
        if not (date is None or date == model.date):
            raise ValueError("cannot set date")
        super().__init__(model, params, date=model.date, **kwargs)

    def _initial_state(self):
        return ()

    def run_to_fill(self, data, times):
        raise RuntimeError


class ClinicalODEModel(ClinicalModel, ODEModel, ABC):
    """
    Base class for clinical models based on ordinary differential equations.
    """
